using System;
using System.IO;			// for StreamReader/StreamWriter I/O to a file
using System.Windows.Forms;	// for DateTimePicker

namespace MyNamespace
{
	/// <summary>
	///		Multiple precision floating-point algorithms for computing Pi = 3.14159...
	///		Developed in Turbo Pascal 5.5.  Converted to Borland C++ 3.1, 4.0 and
	///		5.02 for Windows. Now in C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class MultiPi
	{
		public MultiPi()
		{
			// no constructor logic
		}

		public static void QuRtIvF(MultiF x, MultiF z) // z = x ^ (-1/4), inverse 4th root
		// x is not changed, it is ok for x and this to have the same location in memory
		// but then x will be changed.  Sets muErr True if error
		// Sets muErr True if x = 0
		{
			int sh, saveFMC;
			int mL = 0;
			sh = Math.Min(z.M, Calc.fMC) + 1;
			if (!MultiFD.expanded)  sh += MultiFD.fGuard;	// n guard "digits", no sticky bit
			if (sh > MultiID.muNMax)  sh = MultiID.muNMax;
			if (!Object.ReferenceEquals(x, z))	{ mL = z.M;  MultiFD.ReSizeF(1, z); }
			MultiF t = new MultiF(sh);
			if (sh > 14) {
				MultiF xL = new MultiF(1);
				MultiF y = new MultiF(sh);
				int ps, pl = 7;
				saveFMC = Calc.fMC;
				Calc.fMC = pl+2;
				QuRtIvF(x, t);			// t ~= 1/SqRt(SqRt(x))
				while (pl <= sh)
				{
					pl += pl;
					if (pl > sh)  pl = sh + 1;
					int p;
					for(p = sh + 1; p > pl; p=(p+1)/2)  ;
					pl = p;
					Calc.fMC = pl+2;
					MultiFD.SqF(t, y);
					MultiFD.SqF(y, y);
					if (x.N > pl+2)
					{
						ps = x.N - (pl+2);
						MultiFD.ReducePrecisionF(ps, x, xL);
						MultiFD.MulF(xL, y, y);
					}
					else
					{
						MultiFD.MulF(x, y, y);
					}
					MultiFD.SubF(MultiFD.fOne, y, y);
					MultiFD.MulF(y, t, y);
					MultiFD.RDiv1F(4, y);
					MultiFD.RAddF(y, t); // (1 - x * t^4) * t / 4 + t  // computing x ^ (-1/4)
				}
				xL.V = null;
				Calc.fMC = saveFMC;
				MultiFD.ReSizeF(1, y);
			}
			else
			{
				MultiFD.SqRootF(x, t);
				MultiFD.SqRootF(t, t);
				MultiFD.DivF(MultiFD.fOne, t, t);
			}
			if (!Object.ReferenceEquals(x, z))  MultiFD.ReSizeF(mL, z);
			MultiFD.SetToF(t, z);
		} // QuRtIv

		public static void SqRtIvF(MultiF x, MultiF z) // z = 1 / SqRt(x)
		// x is not changed, it is ok for x and this to have the same location in memory
		// but then x will be changed.  Sets muErr True if error
		// Sets muErr True if x = 0
		{
			int sh, saveFMC;
			int mL = 0;
			sh = Math.Min(z.M, Calc.fMC) + 1;
			if (!MultiFD.expanded)  sh += MultiFD.fGuard; // Four guard "digits", no sticky bit
			if (sh > MultiID.muNMax)  sh = MultiID.muNMax;
			if (!Object.ReferenceEquals(x, z))  { mL = z.M;  MultiFD.ReSizeF(1, z); }
			MultiF t = new MultiF(sh);
			if (sh > 14)
			{
				MultiF xL = new MultiF(1);
				MultiF y = new MultiF(sh);
				int ps, pl = 7;
				saveFMC = Calc.fMC;
				Calc.fMC = pl+2;
				SqRtIvF(x, t);  // t ~= 1/SqRt(x)
				while (pl <= sh)
				{
					pl += pl;
					if (pl > sh)  pl = sh + 1;
					int p;
					for(p = sh + 1; p > pl; p=(p+1)/2)  ;
					pl = p;
					Calc.fMC = pl+2;
					if (x.N > pl+2)
					{
						ps = x.N - (pl+2);
						MultiFD.ReducePrecisionF(ps, x, xL);
						MultiFD.MulF(xL, t, y);
					}
					else
					{
						MultiFD.MulF(x, t, y);
					}
					MultiFD.MulF(y, t, y);
					MultiFD.SubF(MultiFD.fOne, y, y);
					MultiFD.MulF(y, t, y);
					MultiFD.RDiv1F(2, y);
					MultiFD.RAddF(y, t); // (1 - x * t * t) * t / 2 + t  // computing 1/SqRt(x)
				}
				xL.V = null;
				Calc.fMC = saveFMC;
				MultiFD.ReSizeF(1, y);		// release memory
			}
			else
			{
				MultiFD.SqRootF(x, t);
				MultiFD.DivF(MultiFD.fOne, t, t);
			}
			if (!Object.ReferenceEquals(x, z)) MultiFD.ReSizeF(mL, z);		// NMax(mL);
			MultiFD.SetToF(t, z);
		} // SqRtIv

		public static void PiAa(MultiF x) // Compute x = Pi by Borwein algorithm a
		// Uses 3 large "Registers" x, y, and z.  l is small.
		{
			int i, j;
			int it; // Number of iterations(...
			string st;
			double timeAtStart;
			double deltaTime;
			timeAtStart = MultiID.TimerF();
			MultiFD.ReSizeF(Calc.fMC, x);		// 1: working alpha and final output
			MultiF z = new MultiF(Calc.fMC);	// 2: Temp
			MultiF y = new MultiF(Calc.fMC);	// 3: y0, y1, ...
			MultiF l = new MultiF(2);			// -(2**N)
			it = (int)(Math.Log(1024.0 * (Calc.fMC * MultiID.muDMax + 3) / 2788.0 ) / Math.Log(2.0)) + 2;
			Common.DoubleSpace();
			MultiID.Diag("Starting Pi");
			int digits = (Calc.fMC - 1) * MultiID.muDMax + 1;
			st = "  Starting to compute Pi by Borwein algorithm a, " + digits + " decimal digits";
			MultiID.WriteLn(st);
			if (MultiID.muAbort || MultiID.muErr) // if not enough memory
			{
				MultiFD.ReSizeF(1, x);			// clear X
				return;
			}
			MultiFD.ValueF("0.5", out i, x);	// X0 = 1/2 = 0.5
			MultiFD.SetTo1F(2, y);
			SqRtIvF(y, y);						// Y0 = 1/SqRt(2) = SqRt(0.5)
			MultiFD.SetTo1F(-1, l);				// l  = -1
			for (j = 1; j <= it; j++)
			{
				MultiID.MuInterrupt();
				if (MultiID.muErr) break;
				if (MultiID.muAbort)
				{
					MultiID.Diag("Operator abort");
					break;
				}
				MultiID.Diag("Starting Iteration");
				st = "  Start of iteration number " + j + " of " + it;
				MultiID.WriteLn(st);
				// y = (1 - SqRoot(1 - y**2)) / (1 + SqRoot(1 - y**2))
				// y = (SqRtIv(1 - y**2) - 1) / (SqRtIv(1 - y**2) + 1)
				MultiFD.RSqF(y);
				MultiFD.ChSF(y);
				MultiFD.RAdd1F(1, y);
				SqRtIvF(y, y);
				MultiFD.Add1F(y, 1, z);			// z = SqRtIv(1 - y**2) + 1
				MultiFD.RAdd1F(-1, y);			// = SqRtIv(1 - y**2) - 1
				MultiFD.RDivF(z, y);
				// x = (1 + y)**2 * x - 2**N * y
				MultiFD.RAddF(l, l);			// l = -2**1, -2**2, ...
				MultiFD.Add1F(y, 1, z);			// z = (1 + y)
				MultiFD.RSqF(z);
				MultiFD.RMulF(z, x);			// x = (1 + y)**2 * x
				MultiFD.MulF(y, l, z);			// z = - 2**N * y
				MultiFD.RAddF(z, x);
			}
			if (!MultiID.muAbort && !MultiID.muErr)
			{
				MultiID.Diag("All done but...");
				MultiID.WriteLn("  All done but inverting alpha");
				MultiFD.SetTo1F(1, z);			// PI ~= 1 / x
				MultiFD.DivF(z, x, x);
				MultiID.Diag("Pi is done");
				MultiID.WriteLn2("  All done computing Pi");
				deltaTime = MultiID.TimerF() - timeAtStart;
				st = string.Format("PiAa, DT = {0:0.00} sec.  ", deltaTime);
				MultiID.WriteTDLn(st);
			}
			else
			{
				MultiFD.ReSizeF(1, x);	// clear X
			}
		} // PiAa

		public static void PiAb(MultiF x) // Compute x = Pi by Borwein algorithm b
		// Uses 3 large "Registers" x, y, and z.  l is small.
		{
			int i, j;
			int it; // Number of iterations(...
			string st;
			double timeAtStart;
			double deltaTime;
			// get bytes allocated, wait (true) for garbage collection, this does not seem to help
			long totalMemory = GC.GetTotalMemory(true);
			timeAtStart = MultiID.TimerF();
			MultiFD.ReSizeF(Calc.fMC, x);		// 1: working alpha and final output
			MultiF z = new MultiF(Calc.fMC);	// 2: Temp
			MultiF y = new MultiF(Calc.fMC);	// 3: y0, y1, ...
			MultiF l = new MultiF(2);			// -(2**(2 * N + 1))
			it = (int)(Math.Log(1024.0 * (Calc.fMC * MultiID.muDMax + 3) / 2788.0 ) / Math.Log(4.0)) + 1;
			Common.DoubleSpace();
			MultiID.Diag("Starting Pi");
			int digits = (Calc.fMC - 1) * MultiID.muDMax + 1;
			st = "  Starting to compute Pi by Borwein algorithm b, " + digits + " decimal digits";
			MultiID.WriteLn(st);
			if (MultiID.muAbort || MultiID.muErr) // if not enough memory
			{
				MultiFD.ReSizeF(1, x);			// clear X
				return;
			}
			MultiFD.ValueF("0.5", out i, y);	// Y0 = SqRt(2) - 1 = 1/SqRt(0.5) - 1
			SqRtIvF(y, y);
			MultiFD.SetToF(y, x);				// Temp x = SqRt(2)
			MultiFD.RAdd1F(-1, y);
			MultiFD.RMul1F(-4, x);				// X0 = 6 - 4 * SqRt(2);
			MultiFD.RAdd1F(6, x);
			MultiFD.SetTo1F(-2, l);
			for (j = 1; j <= it; j++)
			{
				MultiID.MuInterrupt();
				if (MultiID.muErr) break;
				if (MultiID.muAbort)
				{
					MultiID.Diag("Operator abort");
					break;
				}
				MultiID.Diag("Starting Iteration");
				st = "  Start of iteration number " + j + " of " + it;
				MultiID.WriteLn(st);
				// y = (1 - 4thRoot(1 - y**4)) / (1 + 4thRoot(1 - y**4))
				// y = (Inverse4thRoot(1 - y**4) - 1) / (inverse4thRoot(1 - y**4) + 1)
				MultiFD.RSqF(y);
				MultiFD.RSqF(y);
				MultiFD.ChSF(y);
				MultiFD.RAdd1F(1, y);
				QuRtIvF(y, y);
				MultiFD.Add1F(y, 1, z);			// z = Inverse4thRoot(1 - y**4) + 1
				MultiFD.RAdd1F(-1, y);			// = Inverse4thRoot(1 - y**4) - 1
				MultiFD.RDivF(z, y);
				// x = (1 + y)**4 * x - 2**(2*N + 1) * y * (1 + (1 + y)*y)
				MultiFD.RMul1F(4, l);			// l = -2**3, -2**5, ...
				MultiFD.Add1F(y, 1, z);			// z = (1 + y)
				MultiFD.RSqF(z);
				MultiFD.RSqF(z);
				MultiFD.RMulF(z, x);			// = (1 + y)**4 * x
				MultiFD.Add1F(y, 1, z);			// z = (1 + y)
				MultiFD.RMulF(y, z);
				MultiFD.RAdd1F(1, z);
				MultiFD.RMulF(y, z);
				MultiFD.RMulF(l, z);			// z = - 2**(2*N + 1) * y * (1 + (1 + y)*y
				MultiFD.RAddF(z, x);
			}
			if (!MultiID.muAbort && !MultiID.muErr)
			{
				MultiID.Diag("All done but...");
				MultiID.WriteLn("  All done but inverting alpha");
				MultiFD.SetTo1F(1, z);			// PI ~= 1 / x
				MultiFD.DivF(z, x, x);
				MultiID.Diag("Pi is done");
				MultiID.WriteLn2("  All done computing Pi");
				deltaTime = MultiID.TimerF() - timeAtStart;
				st = string.Format("PiAb, DT = {0:0.00} sec.  ", deltaTime);
				MultiID.WriteTDLn(st);
			}
			else
			{
				MultiFD.ReSizeF(1, x);			// clear x
			}
		} // PiAb

		public static void PiGL(MultiF x, bool SpeedUp)
		// Compute x = Pi by Gauss-Legendre algorithm.
		// See Yasumasa KANADA Computer Centre, University of Tokyo.
		// Uses 4 large "Registers" x, y, z, and t.  l is small.
		// if SpeedUp then speed up SqRootF() with guess, optional, default = True
		{
			int i, j;
			int it;		// Number of iterations(...
			float plG;	// Places the guess to SqRoot is good, in "Super" digits
			float plG0 = (float)((1396.0 / 1024.0) / MultiID.muDMax);
			string st;
			double timeAtStart;
			double deltaTime;
			timeAtStart = MultiID.TimerF();
			MultiFD.ReSizeF(Calc.fMC, x);		// 1: working number and final output
			MultiF t = new MultiF(Calc.fMC);	// 2: Temp
			MultiF z = new MultiF(Calc.fMC);	// 3: Temp
			MultiF y = new MultiF(Calc.fMC);	// 4: y0, y1, ...
			MultiF l = new MultiF(2);			// -(2**N)
			it = (int)(Math.Log(1024.0 * (Calc.fMC * MultiID.muDMax + 3) / 2788.0 ) / Math.Log(2.0)) + 1;
			plG = 0;
			if (SpeedUp)  plG = plG0;
			Common.DoubleSpace();
			MultiID.Diag("Starting Pi");
			int digits = (Calc.fMC - 1) * MultiID.muDMax + 1;
			st = "  Starting to compute Pi by Gauss-Legendre algorithm, " + digits + " decimal digits";
			int tot = 0;
			MultiID.WriteT(st, ref tot);
			if (SpeedUp)
			{
				MultiID.WriteTLn("");
			}
			else
			{
				MultiID.WriteTLn(" without speed up", ref tot);
			}
			MultiFD.SetTo1F(1, x);				// X0 = 1
			MultiFD.SetTo1F(2, y);				// Y0 = 1/SqRt(2) = 0.707...
			SqRtIvF(y, y);
			MultiFD.ValueF("0.25", out i, t);	// T0 = 1/4
			MultiFD.SetTo1F(-1, l);				// L0 = -1
			for (j = 1; j <= it; j++)
			{
				MultiID.MuInterrupt();
				if (MultiID.muErr) break;
				if (MultiID.muAbort)
				{
					MultiID.Diag("Operator abort");
					break;
				}
				MultiID.Diag("Starting Iteration");
				st = "  Start of iteration number " + j + " of " + it;
				MultiID.WriteLn(st);
				MultiFD.SetToF(x, z);		// z = x
				MultiFD.RAddF(y, x);		// x = (x+y) / 2
				MultiFD.RDiv1F(2, x);
				MultiFD.RMulF(z, y);		// y = SqRt(y * z)
				plG += plG;
				MultiFD.SqRootF(y, x, (int)plG, y);
				MultiFD.RSubF(x, z);		// z - x
				MultiFD.RSqF(z);			// (z - x) ** 2
				MultiFD.RMulF(l, z);		// l * (z - x) ** 2
				MultiFD.RAddF(z, t);		// t = t + l * (z - x) ** 2
				MultiFD.RAddF(l, l);		// l = 2 * l
			}
			if (!MultiID.muAbort && !MultiID.muErr)
			{
				MultiID.Diag("All done but...");
				MultiID.WriteLn("  All done but ((x+y)^2) / (4*t)");
				MultiFD.RAddF(y, x);		// PI ~= ((x+y)^2) / (4*t)
				MultiFD.RSqF(x);
				MultiFD.RDivF(t, x);
				MultiFD.ValueF("0.25", out i, l);
				MultiFD.RMulF(l, x);
				MultiID.Diag("Pi is done");
				MultiID.WriteLn2("  All done computing Pi");
				deltaTime = MultiID.TimerF() - timeAtStart;
				st = string.Format("PiGL, DT = {0:0.00} sec.  ", deltaTime);
				MultiID.WriteTDLn(st);
			}
			else
			{
				MultiFD.ReSizeF(1, x);		// clear x
			}
		} // PiGL

		public static void PiAGM(MultiF x, bool SpeedUp)
		// Compute x = Pi by Schoenhage's AGM algorithm.
		// See C++ author = Joerg Arndt		email: arndt@jjj.de
		// the C++ software is online at	http://www.jjj.de/
		// Uses 5 large "Registers" x, a, b, t, and s.
		// if SpeedUp then initialize with first iteration
		{
			int i, j, k;
			int it;			// Number of iterations(...
			string st;
			int pow2;	// 2^(k+1)
			double timeAtStart;
			double deltaTime;
			int n = Calc.fMC;
			timeAtStart = MultiID.TimerF();
			MultiFD.ReSizeF(Calc.fMC, x);		// 1: working number and final output
			MultiF a = new MultiF(Calc.fMC);	// 2: arithmetic mean
			MultiF b = new MultiF(Calc.fMC);	// 3: geometric mean
			MultiF t = new MultiF(Calc.fMC);	// 4: t
			MultiF s = new MultiF(Calc.fMC);	// 5: s
			it = (int)(Math.Log(1024.0 * (Calc.fMC * MultiID.muDMax + 3) / 2788.0 ) / Math.Log(2.0)) + 1;
			Common.DoubleSpace();
			MultiID.Diag("Starting Pi");
			int digits = (Calc.fMC - 1) * MultiID.muDMax + 2;
			st = "  Starting to compute Pi by Schoenhage's AGM algorithm, " + digits + " decimal digits";
			MultiID.WriteLn(st);
			if (MultiID.muAbort || MultiID.muErr)	// if not enough memory
			{
				MultiFD.ReSizeF(1, x);				// clear x
				return;
			}
			if (SpeedUp)
			{
				MultiFD.SetTo1F(2, b);				// b = 2
				SqRtIvF(b, b);						// b = 1/SqRt(2) = SqRt(1/2)	(SqRt inverse function)
				MultiFD.Div1F(b, 2, x);				// x = SqRt(1/2)/2
				MultiFD.SetToF(x, t);				// t = x
				MultiFD.ValueF("0.5", out i, a);	// a = 1/2 = 0.5
				MultiFD.AddF(a, x, a);				// a = sqrt(1/2)/2 + 1/2
				MultiFD.ValueF("0.125", out i, s);	// s = 1/8 = 0.125
				MultiFD.SubF(t, s, t);				// t = sqrt(1/2)/2 - 1/8
				MultiFD.Mul1F(s, 3, s);				// s = 3/8
				MultiFD.AddF(x, s, x);				// x = sqrt(1/2)/2 + 3/8
				MultiFD.Mul1F(t, 2, t);				// t = sqrt(1/2) - 1/4
				k = 1;
			}
			else
			{
				MultiFD.SetTo1F(1, a);				// a = 1
				MultiFD.SetTo1F(1, x);				// x = a^2 = 1
				MultiFD.ValueF("0.5", out i, b);	// b = 1/2 = 0.5
				MultiFD.SetToF(b, t);				// t = 1/2 = 0.5
				k = 0;
				it++;
			}
			long lastPrec = 0;
			for (j = 1;  k < 30;  j++, k++)
			{
				MultiID.MuInterrupt();
				if (MultiID.muErr) break;
				if (MultiID.muAbort)
				{
					MultiID.Diag("Operator abort");
					break;
				}
				MultiID.Diag("Starting Iteration");
				st = "  Start of iteration number " + j + " of " + it;
				MultiID.WriteLn(st);

				MultiFD.AddF(x, b, s);			// s = x + b
				MultiFD.Div1F(s, 4, s);			// s = (x+b) / 4
				MultiFD.SqRootF(b, b);			// b = sqrt(b)		(Full square root)
				MultiFD.AddF(a, b, a);			// a' = a + b
				MultiFD.Div1F(a, 2, a);			// a' = (a+b) / 2
				MultiFD.SqF(a, x);				// x = a'^2			(Full square multiply)
				MultiFD.SubF(x, s, b);			// b = x - s
				MultiFD.Mul1F(b, 2, b);			// b = 2*(x-s)
				MultiFD.SubF(x, b, s);			// s = x - b
				pow2 = 1 << (k+1);				// pow2 = 2^(k+1)
				MultiFD.Mul1F(s, pow2, s);		// s = 2^(k+1) * (x-b)
				MultiFD.SubF(t, s, t);			// t = t - 2^(k+1) * (x-b)

				if (lastPrec >= n / 2) break;
				lastPrec = -s.N - s.C - 2;
			}
			if (!MultiID.muAbort && !MultiID.muErr)
			{
				MultiID.Diag("All done but...");
				MultiID.WriteLn("  All done but calculation pi = 2*a*a/t");

				MultiFD.SqF(a, a);				// a' = a*a
				MultiFD.Mul1F(a, 2, a);			// a' = 2*a*a
				MultiFD.DivF(a, t, x);			// pi = 2*a*a/t

				MultiID.Diag("Pi is done");
				MultiID.WriteLn2("  All done computing Pi");
				deltaTime = MultiID.TimerF() - timeAtStart;
				st = string.Format("PiAGM, DT = {0:0.00} sec.  ", deltaTime);
				MultiID.WriteTDLn(st);
			}
			else
			{
				MultiFD.ReSizeF(1, x);			// clear x
			}
		} // PiAGM
	}
}
