using System;
using System.IO;			// for StreamWriter I/O to a file
using System.Windows.Forms;	// for RichTextBox

namespace MyNamespace
{
	/// <summary>
	///		Extra precision floating-point List algorithms
	///		Developed in TURBO Pascal 5.5 and 6.0, converted to Visual Basic 6
	///		and now in Visual C# .Net
	///
	/// ---------------- *** LEGAL NOTICE: *** ------------------
	///
	/// This program is free software; you can redistribute it and/or modify
	/// it under the terms of the GNU General Public License (GPL) as published
	/// by the Free Software Foundation. cf. the file COPYING.txt.
	///
	/// ------------ *** end of legal notice *** ------------------
	/// </summary>

	public class XPCalLst
	{
		public XPCalLst()
		{
			limFac = 2000;
			InitGammaSeries();
		}

		public static void Expand() // expand the precision for guard digits
		// this may slightly exceed MultiID.muNMax
		{
			if (!MultiFD.expanded)
			{
				MultiFD.expanded = true;
				MultiFD.fMCB = Calc.fMC;	// save basic precision
				Calc.fMC += MultiFD.fGuard;	// n guard "digits"
				MultiFD.mMaxB = MultiID.mMax;
				MultiID.mMax += MultiFD.fGuard + 1;
				if (MultiID.mMax > MultiID.muNMax) MultiID.mMax = MultiID.muNMax;
			}
		}

		public static void RAddN(MultiF x, int zI)  // z = z + x
		// sets muErr true if overflow
		{
			AddN(Common.nList[zI], x, zI);
		}

		public static void RSubN(MultiF x, int zI) // z = z - x
		// sets muErr true if overflow
		{
			SubN(Common.nList[zI], x, zI);
		}

		public static void AddN(MultiF x, MultiF y, int zI) // z = x + y
		// sets muErr true if overflow
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.AddF(x, y, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void SubN(MultiF x, MultiF y, int zI) // z = x - y
		{
			// sets muErr true if overflow
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SubF(x, y, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void SetTo1N(int d1, int zI) // z = d1 % baseM
		// allows a literal d1, d1 an integer, |d1| < baseM, d1 not changed
		{
			MultiF t = new MultiF(2);
			MultiFD.SetTo1F(d1, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void SetToDbN(double db, int zI) // z = db % baseM^4
		// allows a literal db, db not changed, muErr set true if overflow
		{
			MultiF t = new MultiF(4);
			MultiFD.SetToDbF(db, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RAdd1N(int d1, int zI) // z = z + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			MultiF t = new MultiF(2); // temp
			MultiFD.SetTo1F(d1, t);
			RAddN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Add1N(MultiF x, int d1, int zI) // z = x + d1
		// allows literal d1, d1 is not changed, sets muErr true if overflow
		{
			Calc.SetToN(x, zI);
			RAdd1N(d1, zI);
		}

		public static void NewLine(StreamWriter outF, ref int line, ref int page)
		{
			MultiID.WriteLn(null, outF);
			line++;
			if (line == 60)
			{
				if (!Common.quiet) MultiID.WriteLn("Page " + page);
				line = 0;
				MultiID.WriteLn(null, outF);
				MultiID.WriteLn(null, outF, Common.StringOf(27, " ") + "Page " + page);
				MultiID.WriteLn2(null, outF);
				MultiID.WriteLn2(null, outF);
				page++;
			}
		}

		public static void FormatNB(StreamWriter outF, int zI)
		// output z as a formatted set of lines in base baseO
		{
			int tot = 0;
			MultiID.WriteLn(null, outF);
			MultiID.WriteTLn2(null, outF, "Base(" + MultiID.baseO + ")", ref tot, 0);
			MultiID.WriteT(null, outF, Common.nList[zI].Nm + " =", ref tot, 0);
			MultiFD.WriteF(null, outF, Common.nList[zI], ref tot, 0); // WriteF will call WriteFB
			MultiID.WriteLn(null, outF);
		}

		public static void FormatN(StreamWriter outF, int zI)
		// output z as a formatted set of lines and pages
		// z is not modified
		{
			string nm;
			int n;
			int i;
			int j;
			string st;
			string st2;
			int zL;
			string ch;
			int k;
			int did;
			decimal dE;
			int page;
			int line;	// line number of last line printed, 0 - 66
			if (MultiID.baseO != 10)
			{
				FormatNB(outF, zI);
				return;
			}
			nm = Common.nList[zI].Nm;
			n = Common.nList[zI].N;
			st = ((int)(MultiID.baseM + Common.nList[zI].V[n-1])).ToString();
			j = 1;
			do
				j++;
			while ((st[j-1] == '0') && (j != MultiID.muDMax + 1));
			zL = j - 2;
			page = 1;
			MultiID.WriteLn(null, outF);
			MultiID.WriteLn2(null, outF);
			dE = Common.nList[zI].C;
			dE = MultiID.muDMax * (dE + n) - zL - 1;
			if (dE >= 0)
				ch = "+";
			else
				ch = "-";
			st2 = Math.Abs(dE).ToString();
			k = st2.Length;
			if ((k > MultiID.muDpG) && (MultiID.muDpG != 0))
			{
				int jj = k + 1;
				int iL = (k - 1) / MultiID.muDpG;
				for (i = 1; i <= iL; i++)
				{
					jj -= MultiID.muDpG;
					st2 = st2.Substring(0, jj - 1) + "," + st2.Substring(jj - 1);
				}
			}
			MultiID.WriteLn2(null, outF, nm + " = m.n E" + ch + st2 + ", m.n ="); // line may > 80 ch
			line = 6;
			if (Common.nList[zI].S) MultiID.Write(null, outF, "-");
			MultiID.Write(null, outF, st.Substring(j - 1, 1) + ".");
			NewLine(outF, ref line, ref page);
			did = 0;
			k = MultiID.muDMax + 1;
			if (n == 1)
			{
				while (st[k-1] == '0')
					k--;
			}
			if ((j >= k) && (n == 1))
			{
				MultiID.Write(null, outF, "0");
				did = 1; // not in DOD VPCalc
			}
			while (j < k)
			{
				j++;
				MultiID.Write(null, outF, st.Substring(j - 1, 1));
				did++;
				if ((did % 5) == 0) MultiID.Write(null, outF, " ");
			}
			for (i = 2; i <= n; i++)
			{
				st = ((int)(MultiID.baseM + Common.nList[zI].V[n-i])).ToString();
				k = MultiID.muDMax + 1;
				if (i == n)
				{
					while (st[k-1] == '0')
						k--;
				}
				for (j = 2; j <= k; j++)
				{
					MultiID.Write(null, outF, st.Substring(j - 1, 1));
					did++;
					if ((did % 250) == 0)
						NewLine(outF, ref line, ref page);	// new group of 5 lines
					if ((did % 50) == 0)					// new line
						NewLine(outF, ref line, ref page);
					else if ((did % 5) == 0)
					{
						MultiID.Write(null, outF, " ");
					}
				}
			}
			if ((did % 50) != 0)
				NewLine(outF, ref line, ref page);
			dE = Common.nList[zI].C;
			dE = MultiID.muDMax * (dE + n) - zL - 1;
			if (dE >= 0)
				ch = "+";
			else
				ch = "-";
			did++;
			MultiID.Write(null, outF, "E" + ch + st2 + " (" + did + ")");
			NewLine(outF, ref line, ref page);
			while (line != 0)
				NewLine(outF, ref line, ref page);
				// MultiID.Write(null, outF, ((char)26).ToString()); // EOF character (1A hex) not needed
		} // FormatN

		public static void RMul1N(int d1, int zI) // z = z * d1
		// Multi-precision one or two super digit multiply d1 is not changed
		// muErr set true if z overflows
		{
			MultiF t = new MultiF(2);
			MultiFD.SetTo1F(d1, t);
			RMulN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Mul1N(MultiF x, int d1, int zI) // z = x * d1
		// Multi-precision one or two super digit multiply x and d1 are not changed
		// muErr set true if z overflows
		{
			Calc.SetToN(x, zI);
			RMul1N(d1, zI);
		}

		public static void MulN(MultiF x, MultiF y, int zI) // z = x * y
		// x or y or both may have the same address in memory as z
		// x, y are not changed unless they share memory with z
		// sets muErr true if overflow
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.MulF(x, y, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RMulN(MultiF x, int zI) // z = z * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(Common.nList[zI], x, zI);
		}

		public static void SqN(MultiF x, int zI) // z = x * x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow
		{
			MulN(x, x, zI);
		}

		public static void RSqN(int zI) // z = z * z
		// sets muErr true if overflow
		{
			MulN(Common.nList[zI], Common.nList[zI], zI);
		}

		public static void PowIF(int yI, MultiF b, MultiF p,
						bool MFlag, int fMBI)
		// y = b^p % FMB (% only if MFlag is true), p an integer
		{
			int bL;
			int pL;
			int iL;
			bool fini;
			bool signP;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			signP = p.S;
			if (MultiFD.ZTestF(b)) // if b = 0
			{
				Calc.ClearN(yI);
				if (signP)
					MultiID.MuWriteErr("Cannot raise zero to a negative power");
				else if (MultiFD.ZTestF(p)) // p = 0
					SetTo1N(1, yI); // can raise zero to the zero power, 0^0 = 1
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				return;
			}
			fini = false;
			Common.InitNAdd(1, "PowIF-bL", out bL);
			Calc.SetToN(b, bL);
			Common.InitNAdd(1, "PowIF-pL", out pL);
			Calc.SetToN(p, pL);
			Common.nList[pL].S = false;
			SetTo1N(1, yI);
			do
			{
				if (Common.nList[pL].C <= 0)
					iL = (int)(Common.nList[pL].V[Common.nList[pL].C]);
				else
					iL = 0;
				if ((iL % 2) != 0) // Odd(iL)
					MulN(Common.nList[bL], Common.nList[yI], yI);
				if (MFlag) RModN(Common.nList[fMBI], yI);
				RDiv1N(2, pL);
				RIntN(pL);
				if (Common.ZTestN(pL))
					fini = true;
				else
				{
					MulN(Common.nList[bL], Common.nList[bL], bL);
					if (MFlag) RModN(Common.nList[fMBI], bL);
				}
				if (MultiID.muAbort)
				{
					Calc.ClearN(yI);
					fini = true;
				}
			}
			while (!fini);
			Common.LDelete(pL);
			Common.LDelete(bL);
			if (signP) RInvN(yI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(yI);
		} // PowIF

		public static void PowMF(int yI, MultiF b, MultiF p,
						bool MFlag, int fMBI)
		// y = b^p % FMB (% only if MFlag is true)
		{
			int tI;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			if ((p.C >= 0) || MultiFD.ZTestF(b))
			{
				PowIF(yI, b, p, MFlag, fMBI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(yI);
				return;
			}
			if (b.S) // b < 0
			{
				MultiID.MuWriteErr("Cannot raise negative number to a non-integer power");
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.ClearN(yI);
				return;
			}
			Common.InitNAdd(1, "PowMF-tI", out tI);
			LnN(b, tI);
			RMulN(p, tI);
			ExpN(Common.nList[tI], yI);
			if (MFlag) RModN(Common.nList[fMBI], yI);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(yI);
		} // PowMF

		public static void PowN(MultiF b, MultiF p, int zI) // z = b^p
		{
			int fMBI = 0;
			PowMF(zI, b, p, false, fMBI);  // FMB not used
		}

		public static void PowNF(MultiF b, MultiF p, MultiF z) // z = b^p
		{
			int fMBI = 0;
			int zI;
			Common.InitNAdd(1, "Temp-PowNF-zI", out  zI);
			PowMF(zI, b, p, false, fMBI);  // FMB not used
			MultiFD.SetToFS(Common.nList[zI], z);
			Common.LDelete(zI);
		}

		public static void RPowN(MultiF p, int zI) // z = z^p
		{
			PowN(Common.nList[zI], p, zI);
		}

		public static void PowMN(MultiF b, MultiF p, int zI) // z = b^p % FMB
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (fMBI == zI)
			{
				MultiID.WriteTLn("Cannot set Top to PowM(x//  y), FMB cleared");
				Calc.ClearN(fMBI);
				return;
			}
			if (Common.ZTestN(fMBI))
				PowMF(zI, b, p, false, fMBI);
			else
				PowMF(zI, b, p, true, fMBI);
		}

		public static void RPowMN(MultiF p, int zI) // z = z^p % FMB
		{
			PowMN(Common.nList[zI], p, zI);
		}

		public static void RDiv1N(int d1, int zI) // z = z / d1
		// Multi-precision one or two super digit divide, d1 is not changed
		// allows literal d1, muErr set true if d1 = 0
		{
			MultiF t = new MultiF(2);
			MultiFD.SetTo1F(d1, t);
			RDivN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void Div1N(MultiF u, int d1, int zI) // z = u / d1
		// Multi-precision one or two super digit divide, d1 and u not changed
		// u and z may have same address in memory, then u is changed
		// allows literal d1, muErr set true if d1 = 0
		{
			Calc.SetToN(u, zI);
			RDiv1N(d1, zI);
		}

		public static void DivN(MultiF u, MultiF d, int zI) // z = u / d
		// u or d or both may have the same address in memory as z
		// u, d are not changed unless they share memory with z
		// sets muErr true if d = 0
		{
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.DivF(u, d, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RDivN(MultiF d, int zI) // z = z / d
		// d may have the same address in memory as z
		// d is not changed unless it shares memory with z
		// sets muErr true if d = 0
		{
			DivN(Common.nList[zI], d, zI);
		}

		public static void FacNF(MultiF x, int zI, bool MFlag, int fMBI) // z = x! = 1*2*3*...*x
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (x.S)
			{
				MultiID.MuWriteErr("Cannot take factorial of number < zero");
				Calc.ClearN(zI);
				return;
			}
			MultiF xL = new MultiF(4);
			MultiF lim = new MultiF(4);
			MultiFD.IntF(x, xL);
			int j;
			MultiFD.ValueF("4.05997,83000,01705,888E+18", out j, lim);
			if (MultiFD.CompFF(lim, xL) > 0)
			{
				MultiID.MuWriteErr("Number too large for factorial function");
				Calc.ClearN(zI);
				return;
			}
			if (MFlag) MFlag = !Common.ZTestN(fMBI);
			bo = lim | limFac;
			if (!MFlag && (MultiFD.CompFF(lim, xL) > 0))
			{
				xL += 1;					// use the gamma function for large n
				GamNF(xL, zI);
				MultiFD.DoneF(lim);
				MultiFD.DoneF(xL);
				return;
			}
			SetTo1N(1, zI);
			MultiID.nDone = 0;
			MultiID.nToGo = Int32.MaxValue;
			bo = lim | Int32.MaxValue;
			while (MultiFD.CompFF(MultiFD.fOne, xL) > 0)	// while xL > 1
			{
				MulN(xL, Common.nList[zI], zI);  // 99% of processor time of FacF is here
				if (MFlag)
				{
					ModN(Common.nList[zI], Common.nList[fMBI], zI);
					if (MultiFD.ZTestF(Common.nList[zI])) break;
				}
				if (MultiID.muAbort)
				{
					MultiID.WriteLn("Factorial function aborted by operator, " + MultiID.nToGo +
									" multiplies to go");
					Calc.ClearN(zI);
					break;
				}
				xL -= 1;
				if (MultiFD.CompFF(lim, xL) <= 0)
					MultiFD.Get1F(out MultiID.nToGo, xL);
				MultiID.nDone++;
				if (MultiID.statReq2)
				{
					MultiID.statReq2 = false;
					MultiID.WriteLn("Factorial function: (NDone=" + MultiID.nDone + " NToGo=" + MultiID.nToGo + ")");
				}
			}
			MultiID.nToGo = 0;
			MultiFD.DoneF(lim);
			MultiFD.DoneF(xL);
		} // FacNF

		public static void FibNF(MultiF x, int zI) // z = Fibonacci number on Int(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			int saveFMC, i;
			saveFMC = Calc.fMC;
			MultiF a = new MultiF(Calc.fMC);
			MultiF b = new MultiF(Calc.fMC);
			MultiF n = new MultiF(4);
			int tI;
			Common.InitNAdd(1, "FibNF-tI", out tI);	// t = temp on the list
			MultiFD.IntF(x, n);						// n = Int(x)
			MultiFD.ValueF("0.209", out i, a);		// a =  0.209 = decimal digits / n
			MultiFD.MulF(a, n, a);					// a = number of decimal digits in Fib(n)
			MultiFD.Div1F(a, MultiID.muDMax, a);	// a in super digits
			double db;
			MultiFD.GetDbInF(out db, 0, MultiID.muNMax, a);
			Calc.fMC = (int)db + 4;					// fMC in super digits, + for safety
			if (Calc.fMC > saveFMC)
				Calc.fMC = saveFMC;
			n.S = false;							// n = abs(Int(x))
			bo = a | 5;								// a = 5
			MultiFD.RSqRtF(a);						// a = SqRt(5)
			bo = b | a;								// b = SqRt(5)
			b += 1;									// b = 1 + SqRt(5)
			MultiFD.MulF(b, MultiFD.fHalf, b);		// b = (1 + SqRt(5)) / 2
			PowN(b, n, tI);							// t = b ^ n
			MultiFD.DivF(Common.nList[tI], a, b);	// b = b^n / a)
			b += MultiFD.fHalf;						// b = 0.5 + b^n / a
			MultiFD.IntF(b, a);						// a = Int(b)
			if ( x.S && ((n.C != 0) || ((x.V[0] & 1) == 0)) ) // if negative even integer
			{
				a.S = true;							// fib(n) is negative
			}
			Calc.SetToN(a, zI);
			Common.LDelete(tI);
			MultiFD.DoneF(n);
			MultiFD.DoneF(b);
			MultiFD.DoneF(a);
			Calc.fMC = saveFMC;
			Calc.NormN(zI);
		} // FibNF

		public static void FacN(MultiF x, int zI, bool MFlag, int fMBI)
		// z = x! = 1*2*3*...*x % FMB
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			if (MFlag) MFlag = !Common.ZTestN(fMBI);
			if (MFlag) Calc.fMC = Math.Min(2 * Calc.fMC, Common.nMax);
			int tI;
			Common.InitNAdd(1, "FacN-tI", out tI);	// t = temp on the list
			FacNF(x, tI, MFlag, fMBI);
			Calc.SetToN(Common.nList[tI], zI);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // FacN

		public static void FibN(MultiF x, int zI)
		// z = Fibonacci number of Int(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int tI;
			Common.InitNAdd(1, "FibN-tI", out tI);		// t = temp on the list
			FibNF(x, tI);
			Calc.SetToN(Common.nList[tI], zI);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // FacN

		public static void LerchN(MultiF x, MultiF s, MultiF a, int zI) // z = Lerch transendent capitol Phi(x, s, a)
		// x, s, a may have the same address in memory as z
		// x, s, a ares not changed unless it shares memory with z
		// sets muErr true if overflow or error
		//
		// The LerchPhi(x, s, a) function is Psi(x, s, a) = Sum {k = 0, 1, ...} of x^k / |a + k|^s,
		// where any term with a + k = 0 is excluded and |x| <= 1

		{
			if ( (MultiFD.CompFF(MultiFD.fOne, x) > 0) || // if x > 1 or x < -1
				 (MultiFD.CompFF(MultiFD.fMOne, x) < 0) )
			{
				Calc.ClearN(zI);
				MultiID.MuWriteErr("LerchN: Cannot compute LerchPhi for |x| > 1");
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF ap = new MultiF(Calc.fMC);
			MultiF b = new MultiF(Calc.fMC);
			MultiF d = new MultiF(Calc.fMC);
			MultiF del = new MultiF(Calc.fMC);
			MultiF n = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			bo = n | 1;
			bo = ap | a;
			bo = sum | 0;
			bool sEq2 = (MultiFD.CompFF(s, MultiFD.fTwo) == 0); // does s equal 2?
			for (int k = 0; true; k++)		// z = sum{k=0,...}(x^k / (a+k)^s
			{
				if (!MultiFD.ZTestF(ap)) // exclude terms with a + k = 0
				{
					MultiID.MuInterrupt();
					if (MultiID.statReq2)
					{
						Common.SingleSpace();
						MultiID.WriteLn("LerchN: k = " + k);
						MultiID.statReq2 = false;
					}
					if (MultiID.muAbort || MultiID.muErr)
					{ 
						Calc.ClearN(zI);
						break; // exit for
					}
					bool signAp= ap.S;				// save sign of a + k
					ap.S = false;					// ap = absolute value of a + k
					if (sEq2)						// if s equals 2
						bo = d | ap * ap;			//		d = ap^2
					else							// else
						XPCalLst.PowNF(ap, s, d);	//		d = |ap|^s
					ap.S = signAp;					// restore sign of a + k
					bo = del | n / d;
					bo = b | sum;					// b = previous sum
					sum += del;
					if (MultiFD.CompFF(b, sum) == 0)
					{
						Calc.SetToN(sum, zI);
						break;
					}
				}
				n *= x;
				ap += 1;
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LerchN

		public static void GamNF(MultiF x, int zI) // z = gamma(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if ( ((x.C >= 0) && x.S) || MultiFD.ZTestF(x) )
			{
				MultiID.MuWriteErr("Cannot take gamma of integer <= zero");
				Calc.ClearN(zI);
				return;
			}
			MultiF lim = new MultiF(4);
			if (x.C >= 0)							// if x is an integer > 0
			{
				bo = lim | limFac + 1;
				if (MultiFD.CompFF(lim, x) <= 0)	// if x a small integer
				{
					MultiF xL = new MultiF(x.N);
					bo = xL | x - 1;				// use the factorial function for small n
					FacNF(xL, zI, false, 0);
					MultiFD.DoneF(lim);
					MultiFD.DoneF(xL);
					return;
				}
			}
			MultiF limM = new MultiF(4);
			int j;
			MultiFD.ValueF("+4.05997,83000,01705,88908,81E+18", out j, lim);
			MultiFD.ValueF("-4.05997,83000,01705,88908,81E+18", out j, limM);
			if ( (MultiFD.CompFF(lim, x) > 0) || (MultiFD.CompFF(limM, x) < 0) )
			{
				MultiID.MuWriteErr("Number too large for gamma function");
				Calc.ClearN(zI);
				return;
			}
			bo = lim | -5;
			bool reflection = (MultiFD.CompFF(lim, x) < 0);	// if x < -5, use reflection
			MultiFD.DoneF(limM);
			MultiFD.DoneF(lim);
			MultiF xr = new MultiF(Calc.fMC);
			MultiF xx = new MultiF(Calc.fMC);
			if (reflection)								
				bo = xr | 1 - x;						// x = 1 - x, use reflection formula
			else
				bo = xr | x;
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)(Math.Pow(digits, 1.5) / 13.0) + 1;
			int iMax = (int)(0.6 * digits) + 20;
			bo = xx | xr;
			xx -= n;
			xx.S = !xx.S;								// xx = n - xr
			if (xx.S)
				n = 0;
			else
				MultiFD.Get1F(out n, xx);
			bo = xx | (xr + n);							// xx = xr + n
			MultiFD.DoneF(xr);
			MultiF lnG = new MultiF(Calc.fMC);			// lnG = (xx-0.5) * ln(xx) - xx + 0.5 * ln(2*pi)
			bo = lnG | xx;								// lnG = xx
			lnG -= MultiFD.fHalf;						// lnG = xx - 0.5
			int tI;
			Common.InitNAdd(1, "GamNF-tI", out tI);	// t = temp on the list
			Calc.SetToN(xx, tI);
			RLnN(tI);									// t = ln(xx)
			MultiFD.MulF(lnG, Common.nList[tI], lnG);	// lnG = (xx-0.5) * ln(xx)
			lnG -= xx;									// - xx
			int pI = GetPiPtr();
			MulN(MultiFD.fTwo, Common.nList[pI], tI);	// t = 2 * Pi
			RLnN(tI);									// t = ln(2*Pi)
			MulN(MultiFD.fHalf, Common.nList[tI], tI);	// t = 0.5 * ln(2*Pi)
			lnG += Common.nList[tI];
			MultiF j2 = new MultiF(1);
			j2.V[0] = 2;								// j2 = 2
			MultiF j2M = new MultiF(1);
			j2M.V[0] = 1;								// j2M = j2 - 1 = 1
			MultiF xP = new MultiF(Calc.fMC);
			bo = xP | xx;								// xP = xx
			MultiF xSq = new MultiF(Calc.fMC);
			MultiFD.MulF(xx, xx, xSq);					// xSq = xx^2
			int i = Math.Min(iMax, Calc.limBern);
			Common.GenBernI(i);
			MultiF div = new MultiF(Calc.fMC);
			MultiF lnGn = new MultiF(Calc.fMC);
			bool done = false;
			do
			{
				MultiID.MuInterrupt();
				Calc.Bern(j2, tI);				// t = B(j2)
				MultiFD.MulF(j2, j2M, div);		// div = j2 * (j2 - 1)
				MultiFD.MulF(xP, div, div);		// div = j2 * (j2 - 1) * xP
				MultiFD.DivF(Common.nList[tI], div, div);	// div = t / div
				MultiFD.AddF(lnG, div, lnGn);	// lnG new = lnG + new term
				MultiFD.SubF(lnGn, lnG, div);   // div = amount lnG changed
				if (MultiFD.ZTestF(div))
					done = true;
				else
				{
					MultiFD.SetToF(lnGn, lnG);
					MultiFD.MulF(xP, xSq, xP);
					MultiFD.AddF(j2, MultiFD.fTwo, j2);
					MultiFD.AddF(j2M, MultiFD.fTwo, j2M);
				}
			}
			while (!done && !MultiID.muAbort && !MultiID.muErr);
			if (MultiID.diagOn)
			{
				int j2i;
				MultiFD.Get1F(out j2i, j2);
				MultiID.WriteLn("Used Bernoulli number B(" + j2i + ")");
			}
			ExpN(lnG, tI);					// t = Exp(lnG) = gamma(xx)
			bo = div | MultiFD.fOne;
			for (i = 0; i < n; i++)
			{
				MultiFD.Add1F(xx, -1, xx);	// // xx = xx - 1
				MultiFD.MulF(div, xx, div);	// div = div * xx
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr)
				{ 
					Calc.ClearN(tI);
					bo = div | 1;
					break; // exit for
				}
			}
			MultiFD.DoneF(lnGn);
			DivN(Common.nList[tI], div, tI);	// t = gamma(xr) = gamma(xx) / div
			MultiFD.DoneF(div);
			MultiFD.DoneF(xSq);
			MultiFD.DoneF(xP);
			MultiFD.DoneF(j2M);
			MultiFD.DoneF(j2);
			if (reflection)
			{
				int t2I;
				Common.InitNAdd(1, "GamNF-t2I", out t2I);				// t2 = temp on the list
				ModN(x, MultiFD.fTwo, t2I);								// t2 = x mod 2
				if (MultiFD.CompFF(Common.nList[t2I], MultiFD.fMHalf) == 0) // if x mod 2 = -0.5, sin is -1
				{
					Common.nList[tI].S = !Common.nList[tI].S;
				}
				else if (MultiFD.CompFF(Common.nList[t2I], MultiFD.fM1P5) == 0) // // if x mod 2 = -1.5, sin is +1
				{
				}
				else
				{
					bool degModeSave = Calc.degMode;
					Calc.degMode = true;
					RMul1N(180, t2I);									// t2 = (x mod 2) * Pi
					RSinN(t2I);											// t2 = sin(x * Pi)
					RMulN(Common.nList[t2I], tI);						// t = gamma(xr) * sin(x * Pi)
					Calc.degMode = degModeSave;
				}
				DivN(Common.nList[pI], Common.nList[tI], tI);			// t = Pi / t = gamma(x)
				Common.LDelete(t2I);
			}
			Calc.SetToN(Common.nList[tI], zI);
			Calc.NormN(zI);
			Common.LDelete(tI);
			MultiFD.DoneF(lnG);
			MultiFD.DoneF(xx);
		} // GamNF

		public static void BetaNF(MultiF x, int zI) // z = beta(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x))
			{
				Calc.SetToN(MultiFD.fHalf, zI);				// beta(0) = 1/2
				return;
			}
			if ((x.C == 0) & x.S)
			{
				if ((x.V[0] & 1) == 1)
				{
					Calc.ClearN(zI);						// beta(negative odd integer) = 0
					return;
				}
			}
			int saveFMC;
			int saveMMax;
			saveFMC = Calc.fMC;
			saveMMax = MultiID.mMax;
			MultiF xx = new MultiF(Calc.fMC);
			MultiF lim = new MultiF(1);
			// bo = lim | -90000000; // diag
			bo = lim | -313;  // for 400 m -313 is best, for 800 m -477 is best ...
			bool reflection = (MultiFD.CompFF(lim, x) < 0); // if x < lim, use reflection
			// reflection = ((x.V[0] & 3) == 0); // diag
			MultiFD.DoneF(lim);
			if (reflection)								
				bo = xx | 1 - x;							// x = 1 - x, use reflection formula
			else
				bo = xx | x;
			// if (reflection) MultiID.WriteLn("Reflection"); // diag
			MultiF d = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);				// beta = Sum ...
			MultiF sumN = new MultiF(Calc.fMC);
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)((digits / 300.0) * digits + 5.0);
			if (n < 300) n = 300;
			MultiF nF = new MultiF(1);						// n floating
			bo = nF | n;
			int j = 4 * n + 1;
			MultiF iF = new MultiF(1);						// i floating
			MultiF jF = new MultiF(1);						// j floating
			bo = jF | j;
			bo = sum | 0;
			MultiID.ADiag("BetaNF: n = " + n);
			xx.S = !xx.S;									// x = -x for now
			bool sign = false;
			int tI;
			Common.InitNAdd(1, "BetaNF-tI", out tI);		// t = temp on the list
			for (int i = 3; i < j; i += 2)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				sign = !sign;
				bo = iF | i;
				PowN(iF, xx, tI);							// t = i ^ -x
				Common.nList[tI].S = sign;
				MultiFD.AddF(sum, Common.nList[tI], sum);	// sum = sum +/- t
			}
			bo = iF | j;
			PowN(iF, xx, tI);								// t = i ^ -x
			MulN(Common.nList[tI], MultiFD.fHalf, tI); // 1/2J^x
			MultiFD.AddF(sum, Common.nList[tI], sum);		// sum = sum +/- t
			xx.S = !xx.S;									// x = +x now
			MultiF twoToK = new MultiF(Calc.fMC);
			MultiF xK = new MultiF(Calc.fMC);
			MultiF xKP = new MultiF(Calc.fMC);
			MultiF kF = new MultiF(Calc.fMC);
			MultiF term = new MultiF(Calc.fMC);
			bo = twoToK | 4;
			bo = xK | xx;
			xK += 1;
			PowN(iF, xK, tI);								// t = j ^ (x + 1)
			MultiFD.Mul1F(Common.nList[tI], 4, d);			// d = 4 * t
			bo = xK | xx;									// xK = x
			bo = xKP | xK;
			Common.GenBernI(Calc.limBern);
			for (int k = 2; ; k += 2)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				bo = kF | k;
				Calc.Bern(kF, tI);							// t = B(k)
				MultiFD.SubF(twoToK, MultiFD.fOne, term);
				MultiFD.MulF(term, twoToK, term);
				MultiFD.MulF(term, xKP, term);
				MultiFD.AddF(xK, MultiFD.fOne, xK);
				MultiFD.MulF(xK, xKP, xKP);
				if (k > 2)
				{
					MultiFD.MulF(term, xK, term);
					MultiFD.AddF(xK, MultiFD.fOne, xK);
					MultiFD.MulF(xK, xKP, xKP);
				}
				MultiFD.MulF(term, Common.nList[tI], term);
				MultiFD.DivF(term, d, term);
				MultiFD.AddF(sum, term, sumN);				// sum = sum + t
				if (MultiFD.CompFF(sum, sumN) == 0)
				{
					MultiID.ADiag("Used Bernoulli number B(" + k + ")");
					k++; // diag
					break;
				}
				MultiFD.SetToF(sumN, sum);
				MultiFD.MulF(twoToK, MultiFD.fFour, twoToK);	// twoToK = 2^k
				MultiFD.MulF(d, iF, d);							// d = d * j
				MultiFD.MulF(d, iF, d);							// d = d * j
				bo = kF | (k+1);
				MultiFD.MulF(d, kF, d);							// d = d * (k+1)
				bo = kF | (k+2);
				MultiFD.MulF(d, kF, d);							// d = d * (k+2)
			}
			MultiFD.DoneF(term);
			MultiFD.DoneF(kF);
			MultiFD.DoneF(xKP);
			MultiFD.DoneF(xK);
			MultiFD.DoneF(twoToK);
			MultiFD.DoneF(jF);
			MultiFD.DoneF(iF);
			MultiFD.DoneF(nF);
			sum += 1;
			if (reflection)
			{
				ModN(xx, MultiFD.fFour, tI);
				if (MultiFD.CompFF(Common.nList[tI], MultiFD.fOne) == 0) // if x mod 4 = 1, sin is +1
				{
				}
				else if (MultiFD.CompFF(Common.nList[tI], MultiFD.fThree) == 0) // if x mod 4 = 3, sin is -1
				{
					sum.S = !sum.S;
				}
				else
				{
					bool degModeSave = Calc.degMode;
					Calc.degMode = true;
					RMul1N(90, tI);									// t = Pi * x / 2
					RSinN(tI);										// t = sin(Pi * x / 2)
					MultiFD.MulF(Common.nList[tI], sum, sum);		// sum = t * sum
					Calc.degMode = degModeSave;
				}
				int pI = GetPiPtr();
				MultiFD.DivF(MultiFD.fTwo, Common.nList[pI], d);	// 2 / Pi
				PowN(d, xx, tI);									// t = (2 / Pi) ^ x
				MultiFD.MulF(Common.nList[tI], sum, sum);			// sum = t * sum
				GamNF(xx, tI);										// t =  gamma(x)
				MultiFD.MulF(Common.nList[tI], sum, sum);			// sum = t * sum
			}
			Calc.SetToN(sum, zI);
			Calc.NormN(zI);
			Common.LDelete(tI);
			MultiFD.DoneF(sumN);
			MultiFD.DoneF(sum);
			MultiFD.DoneF(d);
			MultiFD.DoneF(xx);
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			if (MultiID.muAbort || MultiID.muErr) Calc.ClearN(zI);
		} // BetaNF

		public static void ErfNF(MultiF x, int zI) // z = erf(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (x.N + x.C > -1)
			{
				// erf(x) = 1 - erfc(x) unless |x| is very small
				ErfCNF(x, zI);
				SubN(MultiFD.fOne, Common.nList[zI], zI);
				return;
			}
			MultiF s = new MultiF(Calc.fMC);
			MultiF m = new MultiF(Calc.fMC);
			MultiF t = new MultiF(Calc.fMC);
			MultiF f = new MultiF(Calc.fMC);
			MultiF c = new MultiF(Calc.fMC);
			MultiF p = new MultiF(Calc.fMC);
			MultiF d = new MultiF(Calc.fMC);
			MultiF xs = new MultiF(Calc.fMC);
			bo = s | 1;			// s = sum
			bo = m | 1;			// m = multiplier
			bo = t | 1;			// t = term
			bo = f | 1;			// f = factorial factor
			bo = c | 3;			// c = constant
			bo = xs | x * x;	// xs = x^2
			for (int j = 1; true; j++)	// sum of (-x^2)^j / (j!*(2*j + 1))
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				m *= xs;			// m = x^2 * m / f
				m /= f;
				m.S = !m.S;			// change sign of m
				bo = t | m / c;		// t = m / c
				bo = p | s;			// p = previous sum
				s += t;				// sum = sum + t
				bo = d | s - p;		// d = delta in sum
				if (MultiFD.ZTestF(d))
				{
					break;
				}
				f += 1;				// f = j
				c += 2;				// c = 2*j + 1
			}
			if (MultiID.muAbort || MultiID.muErr)
			{
				Calc.ClearN(zI);
			}
			else
			{
				s *= 2;				// s = 2 * x * sum
				s *= x;
				int pI = GetPiPtr();
				MultiFD.SqRootF(Common.nList[pI], t);	// t = SqRt(Pi);
				s /= t;				// s = s / SqRt(Pi)
				Calc.SetToN(s, zI);	// final value is s
			}
			MultiFD.DoneF(c);
			MultiFD.DoneF(f);
			MultiFD.DoneF(t);
			MultiFD.DoneF(m);
			MultiFD.DoneF(s);
		} // ErfNF

		public static void ErfCNF(MultiF x, int zI) // z = erfc(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x))				// if x = 0, erfc(x) = 1;
			{
				Calc.SetToN(MultiFD.fOne, zI);
				return;
			}
			MultiF xSq = new MultiF(Calc.fMC);	// if x > 0, erfc(x) = gamq(1/2, x^2)
			bool sign = x.S;					// save sign of x
			MultiFD.MulF(x, x, xSq);
			GamQN(MultiFD.fHalf, xSq, zI);		// because this can clobber x
			if (sign)							// if x was < 0, erfc = 2 - erfc(-x)
			{
				XPCalLst.SubN(MultiFD.fTwo, Common.nList[zI], zI);
			}
			MultiFD.DoneF(xSq);
		} // ErfCNF

		public static void EtaNF(MultiF x, int zI) // z = eta(x), eta(4) = 7*Pi^4/720
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.ZTestF(x))
			{
				Calc.SetToN(MultiFD.fHalf, zI);				// eta(0) = 1/2
				return;
			}
			if ((x.C >= 0) && x.S)
			{
				if ( (x.C > 0) || ((x.V[0] & 1) == 0) )
				{
					Calc.ClearN(zI);						// eta(negative even integer) = 0
					return;
				}
			}
			MultiF xx = new MultiF(Calc.fMC);
			MultiF lim = new MultiF(1);
			bo = lim | -4;
			bool reflection = (MultiFD.CompFF(lim, x) < 0); // if x < -4, use reflection
			MultiFD.DoneF(lim);
			if (reflection)								
				bo = xx | 1 - x;							// x = 1 - x, use reflection formula
			else
				bo = xx | x;
			double digits = Calc.fMC * MultiID.muDMax;
			int n = (int)(1.3 * digits) + 5;
			MultiF nF = new MultiF(2);						// n floating
			bo = nF | n;
			MultiF sum = new MultiF(Calc.fMC);				// eta = Sum ...
			MultiF kF = new MultiF(2);						// k floating = 0
			int k;
			bool sign = false;								// (-1)^(k-1) is negative
			int tI;
			Common.InitNAdd(1, "EtaNF-tI", out tI);			// t = temp on the list
			xx.S = !xx.S; // xx = -x
			MultiF c = new MultiF(Calc.fMC);				// c = C(n, j)
			bo = c | 1;
			MultiF e = new MultiF(Calc.fMC);				// e = Sum C(n, j)
			bo = e | 1;
			MultiF nM = new MultiF(2);						// nM = n - i + 1
			bo = nM | n;
			MultiF iF = new MultiF(2);						// i floating
			bo = iF | 1;
			MultiF d = new MultiF(Calc.fMC);				// d = delta for sum
			for (k = n+n; k > n; k--)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				sign = !sign;
				MultiFD.SetTo1F(k, kF);
				PowN(kF, xx, tI);							// t = k ^ -x
				Common.nList[tI].S = sign;					// * (-1) ^ (k-1)
				MultiFD.MulF(e, Common.nList[tI], d);		// d = e * t
				MultiFD.AddF(sum, d, sum);					// sum = sum + d
				MultiFD.MulF(c, nM, c);						// c = c * (n - i + 1)
				MultiFD.DivF(c, iF, c);						// c = c / i
				MultiFD.AddF(e, c, e);						// e = sum C(n, j)
				MultiFD.SubF(nM, MultiFD.fOne, nM);			// nM = nM -1
				MultiFD.AddF(iF, MultiFD.fOne, iF);			// iF = iF + 1
			}
			MultiFD.DoneF(d);
			int tpI;
			Common.InitNAdd(1, "EtaNF-tpI", out tpI);		// tp = temp on the list
			PowN(MultiFD.fTwo, nF, tpI);					// tp = 2 ^ n			
			MultiFD.DivF(sum, Common.nList[tpI], sum);		// sum = sum / (2 ^ n)
			Common.LDelete(tpI);
			for (k = n; k > 0; k--)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				sign = !sign;
				MultiFD.SetTo1F(k, kF);
				PowN(kF, xx, tI);							// t = k ^ -x
				Common.nList[tI].S = sign;					// * (-1) ^ (k-1)
				MultiFD.AddF(sum, Common.nList[tI], sum);
			}
			MultiFD.DoneF(iF);
			MultiFD.DoneF(nM);
			MultiFD.DoneF(e);
			MultiFD.DoneF(kF);
			if (reflection)
			{
				ModN(x, MultiFD.fFour, tI);
				if (MultiFD.CompFF(Common.nList[tI], MultiFD.fMOne) == 0) // if x mod 4 = -1, sin is -1
				{
					sum.S = !sum.S;
				}
				else if (MultiFD.CompFF(Common.nList[tI], MultiFD.fM3) == 0) // if x mod 4 = -3, sin is +1
				{
				}
				else
				{
					bool degModeSave = Calc.degMode;
					Calc.degMode = true;
					RMul1N(90, tI);								// t = Pi * x / 2
					RSinN(tI);									// t = sin(Pi * x / 2)
					MultiFD.MulF(Common.nList[tI], sum, sum);	// sum = t * sum
					Calc.degMode = degModeSave;
				}
				int pI = GetPiPtr();
				MultiF f = new MultiF(Calc.fMC);				// f = 2 * pi
				MultiFD.MulF(MultiFD.fTwo, Common.nList[pI], f);// f = 2 * Pi
				PowN(f, xx, tI);								// t = (2 * Pi) ^ (x-1)
				MultiFD.MulF(Common.nList[tI], sum, sum);		// sum = t * sum
				xx.S = !xx.S; // xx = +(1-x)
				GamNF(xx, tI); // t = Gamma(1-x)
				MultiFD.MulF(Common.nList[tI], sum, sum);		// sum = t * sum
				PowN(MultiFD.fTwo, xx, tI);						// t = 2 ^ (1-x)
				MultiFD.SubF(MultiFD.fOne, Common.nList[tI], f);// f = 1 - 2^(1-x)
				MultiFD.MulF(f, sum, sum);						// sum = f * sum
				RInvN(tI);										// t = 2 ^ (x-1)
				MultiFD.SubF(MultiFD.fHalf, Common.nList[tI], f);// f = .5 - 2^(x-1)
				MultiFD.DivF(sum, f, sum);						// sum = sum / f
				MultiFD.DoneF(f);
			}
			Calc.SetToN(sum, zI);
			Calc.NormN(zI);
			Common.LDelete(tI);
			MultiFD.DoneF(nF);
			MultiFD.DoneF(xx);
		} // EtaNF

		public static void LambdaNF(MultiF x, int zI) // z = lambda(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.CompFF(x, MultiFD.fOne) == 0)
			{
				MultiID.MuWriteErr("Cannot take lambda(x) for x = one");
				Calc.ClearN(zI);
				return;
			}
			if (MultiFD.ZTestF(x))
			{
				Calc.ClearN(zI);						// lambda(0) = 0
				return;
			}
			if ((x.C >= 0) && x.S)
			{
				if ( (x.C > 0) || ((x.V[0] & 1) == 0) )
				{
					Calc.ClearN(zI);					// lambda(negative even integer) = 0
					return;
				}
			}
			int zetaI, tI;
			Common.InitNAdd(1, "LambdaNF-zetaI", out zetaI);	// zeta = temp on the list
			ZetaNF(x, zetaI);									// lambda(x) = zeta(x) * (1 - 2^x)
			Common.InitNAdd(1, "LambdaNF-t2I", out tI);			// t = temp on the list
			x.S = !x.S;
			PowN(MultiFD.fTwo, x, tI);							// t = 2 ^ (-x)
			x.S = !x.S;
			MultiF f = new MultiF(Calc.fMC);
			MultiFD.SubF(MultiFD.fOne, Common.nList[tI], f);	// f = 1 - 2^(-x)
			MultiFD.MulF(Common.nList[zetaI], f, f);			// f = eta * f
			Calc.SetToN(f, zI);									// zeta(x) = f
			Calc.NormN(zI);
			MultiFD.DoneF(f);
			Common.LDelete(tI);
			Common.LDelete(zetaI);
		} // LambdaNF

		public static void ZetaNF(MultiF x, int zI) // z = zeta(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			if (MultiFD.CompFF(x, MultiFD.fOne) == 0)
			{
				MultiID.MuWriteErr("Cannot take zeta(x) for x = one");
				Calc.ClearN(zI);
				return;
			}
			if (MultiFD.ZTestF(x))
			{
				Calc.SetToN(MultiFD.fMHalf, zI);		// zeta(0) = -1/2
				return;
			}
			if ((x.C >= 0) && x.S)
			{
				if ( (x.C > 0) || ((x.V[0] & 1) == 0) )
				{
					Calc.ClearN(zI);					// zeta(negative even integer) = 0
					return;
				}
			}
			int etaI, tI;
			Common.InitNAdd(1, "ZetaNF-etaI", out etaI);	// eta = temp on the list
			EtaNF(x, etaI);										// Zeta(x) = eta(x) / (1 - 2^(1-x))
			Common.InitNAdd(1, "ZetaNF-t2I", out tI);	// t = temp on the list
			SubN(MultiFD.fOne, x, tI);							// t = 1 - x
			PowN(MultiFD.fTwo, Common.nList[tI], tI);			// t = 2 ^ (1-x)
			MultiF f = new MultiF(Calc.fMC);
			MultiFD.SubF(MultiFD.fOne, Common.nList[tI], f);	// f = 1 - 2^(1-x)
			MultiFD.DivF(Common.nList[etaI], f, f);				// f = eta / f
			Calc.SetToN(f, zI);									// zeta(x) = f
			Calc.NormN(zI);
			MultiFD.DoneF(f);
			Common.LDelete(tI);
			Common.LDelete(etaI);
		} // ZetaNF

		public static void GamNF1(MultiF x, int zI) // z = gamma(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			double intPart;
			if ( ((x.C >= 0) && x.S) || MultiFD.ZTestF(x) )
			{
				MultiID.MuWriteErr("Cannot take gamma of integer <= zero");
				Calc.ClearN(zI);
				return;
			}
			if (x.C >= 0)
			{
				Calc.SetToN(x, zI);
				RSubN(MultiFD.fOne, zI);				// z = x - 1
				FacNF(Common.nList[zI], zI, false, 1);
				return;
			}
			if ((x.N + x.C) <= 2)
				MultiFD.GetDbF(out intPart, x);
			else
				intPart = 1 + (double)Int32.MaxValue;
			if (Math.Abs(intPart) > Int32.MaxValue)				// arbitrary limit
			{
				MultiID.MuWriteErr("Number too large for gamma function");
				Calc.ClearN(zI);
				return;
			}
			MultiF xx = new MultiF(4);
			bo = xx | x - (int)intPart;	// z = fractional part of x
			if (MultiFD.CompFF(MultiFD.fHalf, xx) > 0) // if xx > 0.5
			{
				MultiFD.RAdd1F(-1, xx);
				intPart++;
			}
			if (MultiFD.CompFF(MultiFD.fMHalf, xx) < 0) // if xx < -0.5
			{
				MultiFD.RAdd1F(1, xx);
				intPart--;
			}
			double fracPart;				// fractional part
			xx.C += 4;
			MultiFD.GetDbF(out fracPart, xx);
			fracPart /= Math.Pow(MultiID.baseM, 4);
			xx.C -= 4;
			double oog = GammaSeries(fracPart); // One Over gamma(fracPart)
			Common.nList[zI].V = new int[5];
			Common.nList[zI].M = 4;
			if (oog == 0.0)
			{
				oog = 1.0;
				fracPart = 1.0;
				intPart = intPart - 1;
				if (intPart > 0) MultiFD.RAdd1F(1, xx);
			}
			string st = oog.ToString();
			int ii;
			MultiFD.ValueF(st, out ii, Common.nList[zI]);
			MultiFD.DivF(MultiFD.fOne, Common.nList[zI], Common.nList[zI]); // g
			if (intPart != 0.0)
			{
				MultiID.nToGo = (int)Math.Abs(intPart);
				MultiID.nDone = 0;
				if (intPart > 0)
				{
					for (double i = 1; i <= intPart; i++)
					{
						if (MultiID.muErr)
						{
							MultiID.MuWriteErr("Number too large for gamma function");
							Calc.ClearN(zI);
							break;
						}
						MultiFD.RMulF(xx, Common.nList[zI]);	// g *= xx;
						MultiFD.RAdd1F(1, xx);					// xx += 1;
						MultiID.MuInterrupt();
						if (MultiID.muAbort)
						{
							MultiID.WriteLn("Gamma function aborted by operator, " + (intPart - i) +
											" multiplies to go");
							Calc.ClearN(zI);
							break;
						}
						MultiID.nToGo--;
						MultiID.nDone++;
						if (MultiID.statReq2)
						{
							MultiID.statReq2 = false;
							MultiID.WriteLn("(NDone=" + MultiID.nDone + " NToGo=" + MultiID.nToGo + ")");
						}
					}
				}
				else
				{
					for (double i = 1; i <= -intPart; i++)
					{
						if (MultiID.muErr)
						{
							MultiID.MuWriteErr("Number too large for gamma function");
							Calc.ClearN(zI);
							break;
						}
						if (oog != 1.0)
							MultiFD.RSub1F(1, xx);				// xx -= 1;
						else
							oog = 0.0;
						if (MultiID.ZTestSI(xx))
						{
							MultiID.MuWriteErr("Cannot take gamma of integer <= zero");
							Calc.ClearN(zI);
							return;
						}
						MultiFD.RDivF(xx, Common.nList[zI]);	// g /= xx;
						MultiID.MuInterrupt();
						if (MultiID.muAbort)
						{
							MultiID.WriteLn("Gamma function aborted by operator, " + (-intPart - i) +
											" divides to go");
							Calc.ClearN(zI);
							break;
						}
						MultiID.nToGo--;
						MultiID.nDone++;
						if (MultiID.statReq2)
						{
							MultiID.statReq2 = false;
							MultiID.WriteLn("(NDone=" + MultiID.nDone + " NToGo=" + MultiID.nToGo + ")");
						}
					}
				}
			}
			if (Common.nList[zI].N > 3) Common.nList[zI].V[0] = 0;
			Calc.NormN(zI);
			if (Common.nList[zI].N == 3) // save only the first 16 digits
			{
				double d = 1 + Math.Floor(Math.Log(Common.nList[zI].V[2], 10));
				d = Math.Pow(10, d);
				Common.nList[zI].V[0] = (int)(d * Math.Floor(Common.nList[zI].V[0] / d));
			}
			MultiID.nToGo = 0;
			MultiFD.DoneF(xx);
		} // GamNF1

		public static void PsiNF(MultiF x, int zI) // z = digamma(x)
		// x may have the same address in memory as z
		// x is not changed unless it shares memory with z
		// sets muErr true if overflow or error
		// z must be large enough ... (See FacN)
		{
			 if ( ((x.C >= 0) && x.S) || MultiFD.ZTestF(x) )
			{
				MultiID.MuWriteErr("Cannot take digamma of integer <= zero");
				Calc.ClearN(zI);
				return;
			}
			// numbers are (almost) never too large for Psi
			MultiF lim = new MultiF(1);
			bo = lim | -5;
			bool reflection = (MultiFD.CompFF(lim, x) < 0);	// if x < -5, use reflection
			MultiFD.DoneF(lim);
			MultiF xr = new MultiF(Calc.fMC);
			MultiF xx = new MultiF(Calc.fMC);
			if (reflection)								
				bo = xr | 1 - x;						// x = 1 - x, use reflection formula
			else
				bo = xr | x;
			double digits = Calc.fMC * MultiID.muDMax;
			double dN = (Math.Pow(digits, 1.5) / 13.0) + 1;
			if (dN > Int32.MaxValue)
				dN = Int32.MaxValue;
			int n = (int)dN;
			int iMax = (int)(0.6 * digits) + 20;
			bo = xx | xr;
			xx -= n;
			xx.S = !xx.S;								// xx = n - xr
			if (xx.S)
				n = 0;
			else
				MultiFD.Get1F(out n, xx);
			bo = xx | (xr + n); // xx = xr + n
			MultiFD.DoneF(xr);
			MultiF psi = new MultiF(Calc.fMC);			// psi = ln(xx) - 1 / (2 * xx)
			MultiFD.DivF(MultiFD.fMHalf, xx, psi);		// -0.5 / xx
			int tI;
			Common.InitNAdd(1, "PsiNF-tI", out tI);		// t = temp on the list
			LnN(xx, tI);								// t = ln(xx)
			MultiFD.AddF(psi, Common.nList[tI], psi);	// psi = ln(xx) - 1 / (2 * xx)
			MultiF j2 = new MultiF(1);
			j2.V[0] = 2;								// j2 = 2
			MultiF xP = new MultiF(Calc.fMC);
			MultiF xSq = new MultiF(Calc.fMC);
			MultiFD.MulF(xx, xx, xSq);					// xSq = xx^2
			bo = xP | xSq;								// xP = xSq
			int i = Math.Min(iMax, Calc.limBern);
			Common.GenBernI(i);
			MultiF div = new MultiF(Calc.fMC);
			MultiF psiN = new MultiF(Calc.fMC);
			bool done = false;
			do
			{
				MultiID.MuInterrupt();
				Calc.Bern(j2, tI);							// t = B(j2)
				MultiFD.MulF(xP, j2, div);					// div = j2 * xP
				MultiFD.DivF(Common.nList[tI], div, div);	// div = t / div
				MultiFD.SubF(psi, div, psiN);				// psi new = psi + new term
				MultiFD.SubF(psiN, psi, div);				// div = amount psi changed
				if (MultiFD.ZTestF(div))
					done = true;
				else
				{
					MultiFD.SetToF(psiN, psi);
					MultiFD.MulF(xP, xSq, xP);
					MultiFD.AddF(j2, MultiFD.fTwo, j2);
				}
			}
			while (!done && !MultiID.muAbort && !MultiID.muErr);
			if (MultiID.diagOn)
			{
				int j2i;
				MultiFD.Get1F(out j2i, j2);
				MultiID.WriteLn("Used Bernoulli number B(" + j2i + ")");
			}
			Calc.SetToN(psi, tI);							// t = psi
			bo = div | MultiFD.fOne;
			for (i = 0; i < n; i++)
			{
				MultiFD.Add1F(xx, -1, xx);					// xx = xx - 1
				MultiFD.DivF(MultiFD.fMOne, xx, div);		// div = -1 / xx
				MultiFD.AddF(psi, div, psi);
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr)
				{ 
					Calc.ClearN(tI);
					bo = div | 1;
					break; // exit for
				}
			}
			MultiFD.DoneF(psiN);
			MultiFD.DoneF(div);
			MultiFD.DoneF(xSq);
			MultiFD.DoneF(xP);
			MultiFD.DoneF(j2);
			if (reflection && !MultiID.muAbort && !MultiID.muErr)
			{
				int pI = GetPiPtr();
				int t2I;
				Common.InitNAdd(1, "PsiNF-t2I", out t2I);					// t2 = temp on the list
				ModN(x, MultiFD.fOne, t2I);
				Common.nList[t2I].S = false;
				if (MultiFD.CompFF(Common.nList[t2I], MultiFD.fHalf) == 0)	// if x mod 1 = 0.5, Pi/tan is zero
				{
					Calc.ClearN(tI);
				}
				else if (MultiFD.CompFF(Common.nList[t2I], MultiFD.fP25) == 0)	// if x mod 1 = 0.25, Pi/tan is Pi
				{
					Calc.SetToN(Common.nList[pI], tI);
				}
				else if (MultiFD.CompFF(Common.nList[t2I], MultiFD.fP75) == 0)	// if x mod 1 = 0.75, Pi/tan is -Pi
				{
					Calc.SetToN(Common.nList[pI], tI);
					Common.nList[tI].S = true;
				}
				else
				{
					bool degModeSave = Calc.degMode;
					Calc.degMode = true;
					RMul1N(180, t2I);								// t2 = x * Pi
					RTanN(t2I);										// t2 = tan(x * Pi)
					DivN(Common.nList[pI], Common.nList[t2I], tI);	// t = Pi / tan(x * Pi)
					Calc.degMode = degModeSave;
				}
				MultiFD.AddF(psi, Common.nList[tI], psi);			// psi = psi + Pi / tan(x * Pi)
				Common.LDelete(t2I);
			}
			Calc.SetToN(psi, zI);
			if (MultiID.muAbort || MultiID.muErr) Calc.ClearN(zI);
			Calc.NormN(zI);
			Common.LDelete(tI);
			MultiFD.DoneF(psi);
			MultiFD.DoneF(xx);
		} // PsiNF

		public static void GamN(MultiF x, int zI)
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			GamNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamN

		public static void BetaN(MultiF x, int zI)
		// z = beta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			BetaNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BetaN

		public static void ErfN(MultiF x, int zI)
		// z = error function erf(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			ErfNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // EtaN

		public static void ErfCN(MultiF x, int zI)
		// z = complementary error function erfC(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			ErfCNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ErfCN

		public static void EtaN(MultiF x, int zI)
		// z = eta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			EtaNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // EtaN

		public static void LambdaN(MultiF x, int zI)
		// z = lambda(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			LambdaNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LambdaN

		public static void ZetaN(MultiF x, int zI)
		// z = zeta(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			ZetaNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ZetaN

		public static void GamN1(MultiF x, int zI)
		// z = gamma(x)
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			GamNF1(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamN1

		public static double GammaSeries(double x)
		// Compute 1/gamma(x) using series expansion
		{
			return (((((((((((((((((((((((((c[26]*x + c[25])*x + c[24])*x + c[23])*x + c[22])*x + c[21])*x 
				+ c[20])*x + c[19])*x + c[18])*x + c[17])*x + c[16])*x + c[15])*x + c[14])*x + c[13])*x + c[12])*x + c[11])*x 
				+ c[10])*x + c[9])*x + c[8])*x + c[7])*x + c[6])*x + c[5])*x + c[4])*x + c[3])*x + c[2])*x + c[1])*x;
		}

		public static void InitGammaSeries()
		// Initialize the gamma series
		{
			// Series expansion for 1/gamma(z) = Sum k=1,16 ck*z^k
			// from Handbook of Math Functions, USDofC
			// c1 thru c 26 =
			c = new double[27];
			c[00] = 0.0;
			c[01] = +1.0000000000000000;
			c[02] = +0.5772156649015329;
			c[03] = -0.6558780715202538;
			c[04] = -0.0420026350340952;
			c[05] = +0.1665386113822915;
			c[06] = -0.0421977345555443;
			c[07] = -0.0096219715278770;
			c[08] = +0.0072189432466630;
			c[09] = -0.0011651675918591;
			c[10] = -0.0002152416741149;
			c[11] = +0.0001280502823882;
			c[12] = -0.0000201348547807;
			c[13] = -0.0000012504934821;
			c[14] = +0.0000011330272320;
			c[15] = -0.0000002056338417;
			c[16] = +0.0000000061160950;
			c[17] = +0.0000000050020075;
			c[18] = -0.0000000011812746;
			c[19] = +0.0000000001043427;
			c[20] = +0.0000000000077823;
			c[21] = -0.0000000000036968;
			c[22] = +0.0000000000005100;
			c[23] = -0.0000000000000206;
			c[24] = -0.0000000000000054;
			c[25] = +0.0000000000000014;
			c[26] = +0.0000000000000001;
		}

		public static void PsiN(MultiF x, int zI)
		// z = psi(x) = gamma'(x) / gamma(x) = digamma function
		// x may have the same address in memory as zI
		// x is not changed unless it shares memory with zI
		// sets muErr true if overflow or error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			PsiNF(x, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // PsiN

		public static void RFacN(int zI) // z = z! = 1*2*3*...*x
		// sets muErr true if overflow or error
		{
			int fMBI = 1;
			FacN(Common.nList[zI], zI, false, fMBI);
		}

		public static void RFibN(int zI) // z = Fibonacci number Int(z)
		// sets muErr true if overflow or error
		{
			FibN(Common.nList[zI], zI);
		}

		public static void RFacMBN(int zI) // z = z! = 1*2*3*...*x % FMB
		// sets muErr true if overflow or error
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (fMBI == zI)
			{
				MultiID.WriteTLn("Cannot set Top to FacM(x), FMB cleared");
				Calc.ClearN(fMBI);
				return;
			}
			FacN(Common.nList[zI], zI, true, fMBI);
		}

		public static void RBetaN(int zI) // z = beta(z)
		// sets muErr true if overflow or error
		{
			BetaN(Common.nList[zI], zI);
		}

		public static void RErfN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			ErfN(Common.nList[zI], zI);
		}

		public static void RErfCN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			ErfCN(Common.nList[zI], zI);
		}

		public static void REtaN(int zI) // z = eta(z)
		// sets muErr true if overflow or error
		{
			EtaN(Common.nList[zI], zI);
		}

		public static void RLambdaN(int zI) // z = lambda(z)
		// sets muErr true if overflow or error
		{
			LambdaN(Common.nList[zI], zI);
		}

		public static void RZetaN(int zI) // z = zeta(z)
		// sets muErr true if overflow or error
		{
			ZetaN(Common.nList[zI], zI);
		}

		public static void RGamN(int zI) // z = gamma(z)
		// sets muErr true if overflow or error
		{
			GamN(Common.nList[zI], zI);
		}

		public static void RGamN1(int zI) // z = gamma(z)
		// sets muErr true if overflow or error
		{
			GamN1(Common.nList[zI], zI);
		}

		public static void RPsiN(int zI) // z = psi(z)
		// sets muErr true if overflow or error
		{
			PsiN(Common.nList[zI], zI);
		}

		public static void SqRootN(MultiF x, int zI) // z = SqRt(x)
		// x is not changed, it is ok for x and z to have the same location in memory
		// but then x will be changed.  Sets muErr true if error
		{
			MultiF t = new MultiF(Calc.fMC); // expand done by SqRoorF
			MultiFD.SqRootF(x, t);
			Calc.SetToN(t, zI);
			MultiFD.DoneF(t);
		}

		public static void RSqRtN(int zI) // z = SqRt(z)
		// sets muErr true if error
		{
			SqRootN(Common.nList[zI], zI);
		}

		public static void InvN(MultiF x, int zI) // z = 1 / x
		{
			MultiF t = new MultiF(1);
			MultiFD.SetTo1F(1, t);
			DivN(t, x, zI);  // z will be resized if not large enough
			MultiFD.DoneF(t);
		}

		public static void RInvN(int zI) // z = 1 / z
		{
			InvN(Common.nList[zI], zI);
		}

		public static int GetPiPtr() // returns a pointer to pi on list
		// puts it on the list if not there
		{
			int pI;
			Common.LFind("Pi", out pI);
			if (pI != 0)
			{
				if (Common.nList[pI].N >= Calc.fMC)
				{
					if (Common.nList[pI].N > Calc.fMC)
						Calc.NormN(pI); // reduce precision
					return pI;
				}
			}
			if (pI == 0)
				Common.InitNAdd(1, "Pi", out pI);
			if (!Calc.ReadFile("Pi.XPN", pI))
			{
				MultiPi.PiAb(Common.nList[pI]);	// Compute x = Pi by Borwein algorithm b
			}
			if (Common.nList[pI].N != Calc.fMC)	// assumes pi does not have long strings of zeros
				MultiID.WriteLn("Error in Pi, N = " +
						(Common.nList[pI].N * MultiID.muDMax - 3) + ", M = " + (Calc.fMC * MultiID.muDMax - 3));
			return pI;
		} // GetPiPtr

		public static void SetToPiN(int zI) // compute z = pi
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int xI = GetPiPtr();
			Calc.SetToII(xI, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
		}

		public static int GetLn10Ptr() // returns a pointer to Ln10 on list
		// puts it on the list if not there
		{
			int ln10;
			Common.LFind("Ln10", out ln10);
			if (ln10 != 0)
			{
				if (Common.nList[ln10].N >= Calc.fMC)
				{
					if (Common.nList[ln10].N > Calc.fMC)
						Calc.NormN(ln10);		// reduce precision
					return ln10;
				}
			}
			if (ln10 == 0)
				Common.InitNAdd(1, "Ln10", out ln10);
			if (!Calc.ReadFile("Ln10.XPN", ln10))
			{
				MultiF ten = new MultiF(1);
				MultiFD.SetTo1F(1, ten);
				MultiFD.RDiv1F(10, ten);
				MLnLN(ten, false, ln10);	// take Ln of 0.1
				Common.ChSN(ln10);
				MultiFD.DoneF(ten);
			}
			if (Common.nList[ln10].N != Calc.fMC)
				MultiID.WriteLn("Error in Ln10, fMC = " + Calc.fMC);
			return ln10;
		} // GetLn10Ptr

		public static void SetToLn10N(int zI) // compute z = Ln10
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int xI = GetLn10Ptr();
			Calc.SetToII(xI, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
		}

		public static void SetFMBN(int zI) // set Modulo base FMB = z
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			Calc.SetToII(zI, fMBI);
		}

		public static void ModFMBN(MultiF x, int zI) // z = x % FMB
		{
			int fMBI;
			Common.LFindAdd("FMB", out fMBI);
			if (!Common.ZTestN(fMBI)) RModN(Common.nList[fMBI], zI);
		}

		public static void RModFMBN(int zI) // z = z % FMB
		{
			ModFMBN(Common.nList[zI], zI);
		}

		public static void SSinN(MultiF x, int zI)  // z = Sin(x), small x
		// Sin(x) = x - x^3 / 3! + x^5 / 5! ...
		{
			bool fini;
			int j;
			bool sign;
			int saveFMC;
			int saveMMax;
			int currFMC;
			if (x.C < (-Int32.MaxValue / 3 - Int16.MaxValue))
			{
				Calc.SetToN(x, zI);
				if (Calc.degMode) RToRadN(zI);
				return;
			}
			saveFMC = Calc.fMC;
			saveMMax = MultiID.mMax;
			currFMC = Calc.fMC;
			MultiF nn = new MultiF(2);
			MultiF term = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			sign = x.S;
			MultiFD.SetToF(x, term);	// term	= |x|
			term.S = false;
			MultiFD.SetTo1F(1, nn);		// nn	= 1
			MultiFD.SqF(term, g);		// g	= -Sq(term)
			MultiFD.ChSF(g);
			MultiFD.SetToF(term, sum);	// sum	= term
			fini = false;
			MultiID.nDone = 1;
			while (!fini)
			{
				MultiFD.RMulF(g, term);		// term = term * g / (nn + 1) / (nn + 2)
				MultiFD.RAdd1F(1, nn);		// nn++
				MultiFD.RDivF(nn, term);
				MultiFD.RAdd1F(1, nn);		// nn++
				MultiFD.RDivF(nn, term);
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				long iL = Calc.fMC - sum.N - sum.C + term.N + term.C;
				if (iL > Int32.MaxValue) iL = Int32.MaxValue;
				if (iL < Int32.MinValue) iL = Int32.MinValue;
				MultiID.nToGo = (int)iL;
				MultiID.nDone++;
				if ((MultiID.nToGo < 0) || (MultiFD.ZTestF(term)) || (MultiID.muErr) || (MultiID.muAbort))
					fini = true;
				else
				{
					MultiFD.RAddF(term, sum); // sum  = sum + term
					Calc.fMC = MultiID.nToGo + 1;
					if (Calc.fMC < currFMC)
					{
						currFMC = Calc.fMC;
						j = term.N - Calc.fMC;
						if (j > 0) MultiFD.ShiftR(j, term);
						j = g.N - Calc.fMC;
						if (j > 0) MultiFD.ShiftR(j, g);
					}
					else
					{
						Calc.fMC = currFMC;
					}
				}
			}
			MultiID.nToGo = 0;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToN(sum, zI);
			MultiFD.DoneF(sum);
			MultiFD.DoneF(g);
			MultiFD.DoneF(term);
			MultiFD.DoneF(nn);
			if ((Common.nList[zI].N > -Common.nList[zI].C) && (Common.nList[zI].N > 1))
				SetTo1N(1, zI);			// if |Sin(x)| > 1 then set to 1
			Common.nList[zI].S = sign;
		} // SSinN

		public static void ReportBest(int j) // compute + report best number of iterations
		// j = number of iterations used.
		// ti0 = Base time, ti1 = Delta time of iterations, ti2 = delta time to evaluate series
			{
			Calc.ti1 = MultiID.TimerF() - Calc.ti0 + Calc.ti1;
			if ((Calc.ti1 > Calc.tic) && (Calc.ti2 > Calc.tic) && !MultiID.muErr && !MultiID.muAbort)
			{
				Calc.ti0 = j * Math.Sqrt((Calc.ti2 - Calc.tic) / (Calc.ti1 + Calc.tic));	// Min
				Calc.ti2 = j * Math.Sqrt((Calc.ti2 + Calc.tic) / (Calc.ti1 - Calc.tic));	// Max
				Calc.ti1 = (Calc.ti0 + Calc.ti2) / 2;										// Ave
				Calc.ti2 = Calc.ti1 - Calc.ti0;												// Del
				MultiID.WriteLn("Best  NN = " + Calc.ti1.ToString("0.000") + " +/- " + Calc.ti2.ToString("0.000"));
			}
		}

		public static void MSinN(MultiF x, int zI)  // z = Sin(x), medium small x
		// uses Sin(x) = Sin(x/3) * (3 - 4 * Sq(Sin(x/3))) nn times
		{
			int tI;
			int nn;
			double d;
			bool sign;
			int i;
			int j;
			int k;
			sign = x.S;
			MultiF e = new MultiF(2);
			Common.InitNAdd(1, "MSinN-tI", out tI);
			MultiF n3 = new MultiF(1);
//			d = 2.7242 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.30700581817;  // PolyFit
			d = 2.7242 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.5;			// PolyFit
			Common.LFind("MSinNN", out nn);
			if (nn != 0) MultiFD.GetDbF(out d, Common.nList[nn]);
			j = (int)Math.Round(d);
			if (MultiID.diagOn)
				MultiID.WriteLn("MSin: NN = " + d.ToString("0.000"));
			MultiFD.SetTo1F(3, n3);
			Calc.SetToN(x, tI);
			if (Calc.degMode) RToRadN(tI);
			Common.nList[tI].S = false;
			MultiFD.SetTo1F(2, e);
			for (i = 1; i <= j; i++)  // e = 2 / 3^nn
				MultiFD.RDivF(n3, e);
			MultiID.nDone = 0;
			k = j;
			MultiID.nToGo = k;
			j = 0;
			Calc.ti0 = MultiID.TimerF();
			while (MultiFD.CompFF(e, Common.nList[tI]) > 0)	// t = x / 3^nn, t < e
			{
				RDivN(n3, tI);
				j++;
				if ((MultiID.muErr || MultiID.muAbort)) SetTo1N(0, tI);
				MultiID.nDone++;
				MultiID.nToGo = k - MultiID.nDone;
			}
			Calc.ti1 = MultiID.TimerF() - Calc.ti0;
			if (MultiID.diagOn) MultiID.WriteLn("MSin: NN = " + j);
			Calc.ti0 = MultiID.TimerF();
			SSinN(Common.nList[tI], zI); // S0 = Sin(t)
			d = MultiID.TimerF();
			Calc.ti2 = d - Calc.ti0;
			Calc.ti0 = d;
			for (i = 1; i <= j; i++)	// Si = Si-1 * (3 - Sq(2 * Si-1)) i = 1, ..., nn
			{
				AddN(Common.nList[zI], Common.nList[zI], tI);
				RSqN(tI);
				Common.ChSN(tI);			//  HJS 1992-11-28
				RAddN(n3, tI);
				RMulN(Common.nList[tI], zI);
				MultiID.nToGo = j - i;
				MultiID.nDone = i;
			}
			if (MultiID.diagOn) ReportBest(j);
			MultiFD.DoneF(n3); // Sin(x) = Snn
			Common.LDelete(tI);
			MultiFD.DoneF(e);
			if ((Common.nList[zI].N > -Common.nList[zI].C) && (Common.nList[zI].N > 1))
				SetTo1N(1, zI); // if |Sin(x)| > 1 then set to 1
			Common.nList[zI].S = sign;
		} // MSinN

		public static void SinQuadN(MultiF x, out int qu, int zI)
		// z = Sin(x), qu = Quadrant (1..4), returned
		{
			int qM4;
			int qu1;
			int pI;
			int piO2;
			bool sign;
			int i;
			long d;
			int saveFR;
			bool onYAxis;
			qu = 0; // default value
			// Common.ALet("Inv(3) + 2"); Common.ALet("Sin = Inv(6)"); // for test only
			d = x.N;
			d += x.C;
			if (!MultiFD.expanded) MultiFD.fMCB = Calc.fMC;
			if (d > MultiFD.fMCB)
			{
				MultiID.MuWriteErr("Sin: Angle too large");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF t = new MultiF(Calc.fMC);
			Common.InitNAdd(1, "SinQuadN-piO2", out piO2);
			MultiF q = new MultiF(Calc.fMC);
			Common.InitNAdd(1, "SinQuadN-qM4", out qM4);
			sign = x.S;
			MultiFD.SetToF(x, t);
			t.S = false;
			saveFR = MultiFD.fRound;
			MultiFD.fRound = 0;
			if (Calc.degMode)
			{
				SetTo1N(90, piO2);
				MultiFD.IntF(t, q);
				MultiFD.FracF(t, t);
				long iL = q.N + q.C;
				if (iL > Int32.MaxValue) iL = Int32.MaxValue;
				if (iL < Int32.MinValue) iL = Int32.MinValue;
				i = (int)iL;
				MultiF t1 = new MultiF(i);
				MultiF t2 = new MultiF(i);
				MultiFD.SetToF(q, t1);
				MultiFD.ShiftL(t1.C, t1);
				MultiID.RDivQRSI(Common.nList[piO2], t2, t1);
				Calc.SetToNM(t1, q);
				MultiFD.RAddF(t2, t);
				MultiFD.DoneF(t2);
				MultiFD.DoneF(t1);
			}
			else
			{
				pI = GetPiPtr();
				Div1N(Common.nList[pI], 2, piO2);
				MultiFD.DivF(t, Common.nList[piO2], q);
				MultiFD.RIntF(q);
				MulN(q, Common.nList[piO2], qM4);
				MultiFD.RSubF(Common.nList[qM4], t);
			}
			SetTo1N(4, qM4);
			ModN(q, Common.nList[qM4], qM4);
			MultiFD.Get1F(out qu1, Common.nList[qM4]);
			qu = (int)qu1;
			qu++;
			onYAxis = false;
			if ((qu == 2) || (qu == 4))
			{
				if (MultiFD.ZTestF(t))
					onYAxis = true;
				else
				{
					MultiFD.SubF(Common.nList[piO2], t, t);
				}
			}
			if ((qu == 3) || (qu == 4)) sign = !sign;
			Common.LDelete(qM4);
			MultiFD.DoneF(q);
			Common.LDelete(piO2);
			MultiFD.fRound = saveFR;
			if (MultiFD.ZTestF(t))
			{
				Calc.ClearN(zI);
				if (onYAxis)
				{
					Common.nList[zI].V[0] = 1;
				}
			}
			else
			{
				MSinN(t, zI);
			}
			if (!Common.ZTestN(zI)) Common.nList[zI].S = sign;
			MultiFD.DoneF(t);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // SinQuadN

		public static void SinN(MultiF x, int zI)  // z = Sin(x)
		{
			int qu;
			SinQuadN(x, out qu, zI);
		}

		public static void RSinN(int zI) // z = Sin(z)
		{
			SinN(Common.nList[zI], zI);
		}

		public static void CosN(MultiF x, int zI)  // z = Cos(x)
		// Cos(x) = Sin(x + pi/2)
		{
			int pI;
			int piO2;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			if (Calc.degMode)
				Add1N(x, 90, zI);
			else
			{
				pI = GetPiPtr();
				Common.InitNAdd(1, "CosN-piO2", out piO2);
				Div1N(Common.nList[pI], 2, piO2);
				AddN(x, Common.nList[piO2], zI);
				Common.LDelete(piO2);
			}
			RSinN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // CosN

		public static void RCosN(int zI) // z = Cos(z)
		{
			CosN(Common.nList[zI], zI);
		}

		public static void TanN(MultiF x, int zI)  // z = Tan(x)
		// Tan(x) = Sin(x) / SqRt(1 - Sq(Sin(x))
		// and change sign of Tan(x) if in 2nd or 3rd quadrant
		{
			int qu;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			if (x.C < (-Int32.MaxValue / 3 - Int16.MaxValue))
			{
				Calc.SetToN(x, zI);
				if (Calc.degMode) RToRadN(zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
				return;
			}
			SinQuadN(x, out qu, zI);
			if (Common.ZTestN(zI))
			{
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				return;
			}
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SqF(Common.nList[zI], t);
			t.S = true;
			MultiFD.RAdd1F(1, t);  // 1 - Sin(x)^2
			if (MultiFD.ZTestF(t))
			{
				MultiID.MuWriteErr("Tan(x) error: x equivalent to +/- 90 Deg.");
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.ClearN(zI);
			}
			else
			{
				MultiFD.RSqRtF(t);
				RDivN(t, zI);
				if ((qu == 2) || (qu == 3)) Common.nList[zI].S = !Common.nList[zI].S;
			}
			MultiFD.DoneF(t);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // TanN

		public static void RTanN(int zI) // z = Tan(z)
		{
			TanN(Common.nList[zI], zI);
		}

		public static void SATanN(MultiF x, int zI)  // z = ATan(x), small x
		// ATan(x) = x - x^3 / 3 + x^5 / 5 ...
		{
			bool fini;
			bool sign;
			int j;
			int saveFMC;
			int saveMMax;
			int currFMC;
			if (x.C < (-Int32.MaxValue / 3 - Int16.MaxValue))
			{
				Calc.SetToN(x, zI);
				if (Calc.degMode) RToDegN(zI);
				return;
			}
			saveFMC = Calc.fMC;
			saveMMax = MultiID.mMax;
			currFMC = Calc.fMC;
			MultiF nn = new MultiF(2);
			MultiF term = new MultiF(Calc.fMC);
			MultiF bTerm = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			sign = x.S;
			MultiFD.SetToF(x, bTerm);				// bTerm = |x|
			bTerm.S = false;
			MultiFD.SetTo1F(1, nn);					// nn	= 1
			MultiFD.SqF(bTerm, g);					// g = -Sq(bTerm)
			MultiFD.ChSF(g);
			MultiFD.SetToF(bTerm, sum);				// sum  = bTerm
			fini = false;
			MultiID.nDone = 1;
			while (!fini)
			{
				MultiFD.RMulF(g, bTerm);			// bTerm = bTerm * g
				MultiFD.RAdd1F(2, nn);				// nn	= nn + 2
				MultiFD.DivF(bTerm, nn, term);		// term = bTerm / nn
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				long iL = Calc.fMC - sum.N - sum.C + term.N + term.C;
				if (iL > Int32.MaxValue) iL = Int32.MaxValue;
				if (iL < Int32.MinValue) iL = Int32.MinValue;
				MultiID.nToGo = (int)iL;
				MultiID.nDone++;
				if ((MultiID.nToGo < 0) || (MultiFD.ZTestF(term)) || (MultiID.muErr) || (MultiID.muAbort))
					fini = true;
				else
				{
					MultiFD.RAddF(term, sum);		// sum  = sum + term
					Calc.fMC = MultiID.nToGo + 1;
					if (Calc.fMC < currFMC)
					{
						currFMC = Calc.fMC;
						j = term.N - Calc.fMC;
						if (j > 0) MultiFD.ShiftR(j, term);
						j = g.N - Calc.fMC;
						if (j > 0) MultiFD.ShiftR(j, g);
					}
					else
					{
						Calc.fMC = currFMC;
					}
				}
			}
			MultiID.nToGo = 0;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToN(sum, zI);
			MultiFD.DoneF(sum);
			MultiFD.DoneF(g);
			MultiFD.DoneF(bTerm);
			MultiFD.DoneF(term);
			MultiFD.DoneF(nn);
			Common.nList[zI].S = sign;
		} // SATanN

		public static void MATanN(MultiF x, int zI)  // z = ATan(x), medium small x
		// uses Tan(x/2) = Tan(x) / (1 + SqRt(1 + Sq(Tan(x)), nn times
		{
			int nn;
			double d;
			bool sign;
			int i;
			int j;
			int k;
			sign = x.S;
//			d = 1.5832 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.39892668084;  // PolyFit
			d = 1.5832 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.22;			// PolyFit
			Common.LFind("MATanNN", out nn);
			if (nn != 0) MultiFD.GetDbF(out d, Common.nList[nn]);
			j = (int)Math.Round(d);
			if (MultiID.diagOn) MultiID.WriteLn("MATan: NN = " + d.ToString("0.000"));
			MultiF e = new MultiF(2);
			MultiF f = new MultiF(2);
			MultiFD.SetTo1F(15, e);			// e = 1.5
			MultiFD.RDiv1F(10, e);
			for (i = 1; i <= j; i++)		// e = e / (1 + SqRt(1 + Sq(e))), nn times
			{
				MultiFD.SqF(e, f);
				MultiFD.RAdd1F(1, f);
				MultiFD.RSqRtF(f);
				MultiFD.RAdd1F(1, f);
				MultiFD.RDivF(f, e);
			}
			MultiFD.DoneF(f);
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SetToF(x, t);
			t.S = false;
			MultiID.nDone = 0;
			k = j;
			MultiID.nToGo = k;
			j = 0;
			Calc.ti0 = MultiID.TimerF();
			while (MultiFD.CompFF(e, t) > 0)  // t = t / (1 + SqRt(1 + Sq(t))), until t < e
			{
				SqN(t, zI);
				RAdd1N(1, zI);
				RSqRtN(zI);
				RAdd1N(1, zI);
				MultiFD.RDivF(Common.nList[zI], t);
				j++;
				if (MultiID.muErr || MultiID.muAbort) MultiFD.SetTo1F(0, t);
				MultiID.nDone++;
				MultiID.nToGo = k - MultiID.nDone;
			}
			Calc.ti1 = MultiID.TimerF() - Calc.ti0;
			if (MultiID.diagOn)
				MultiID.WriteLn("MATan: NN = " + j);
			Calc.ti0 = MultiID.TimerF();
			SATanN(t, zI);					// A0 = ATan(t)
			d = MultiID.TimerF();
			Calc.ti2 = d - Calc.ti0;
			Calc.ti0 = d;
			MultiFD.DoneF(t);
			MultiFD.DoneF(e);
			for (i = 1; i <= j; i++)		// Ai = 2 * Ai-1 i = 1, ..., nn
			{
				RMul1N(2, zI);
				MultiID.nToGo = j - i;
				MultiID.nDone = i;
			}
			if (MultiID.diagOn) ReportBest(j);
			Common.nList[zI].S = sign;
			if (Calc.degMode) RToDegN(zI);
		} // MATanN

		public static void ATanN(MultiF x, int zI)  // z = ATan(x)
		{
			int pI;
			int piO2;
			bool sign;
			bool XGT1;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF t = new MultiF(Calc.fMC);
			sign = x.S;
			MultiFD.SetToF(x, t);
			t.S = false;
			XGT1 = false;
			if (MultiFD.CompFF(MultiFD.fOne, t) > 0)
			{
				MultiFD.DivF(MultiFD.fOne, t, t);
				XGT1 = true;
			}
			MATanN(t, zI);
			MultiFD.DoneF(t);
			if (XGT1)
			{
				if (Calc.degMode)
				{
					Common.ChSN(zI);
					RAdd1N(90, zI);
				}
				else
				{
					pI = GetPiPtr();
					Common.InitNAdd(1, "ATanN-piO2", out piO2);
					Div1N(Common.nList[pI], 2, piO2);
					SubN(Common.nList[piO2], Common.nList[zI], zI);
					Common.LDelete(piO2);
				}
			}
			Common.nList[zI].S = sign;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ATanN

		public static void RATanN(int zI) // z = ATan(z)
		{
			ATanN(Common.nList[zI], zI);
		}

		public static void ATan2N(MultiF y, MultiF x, int zI)  // z = ATan(y / x)
		// ComPFute Arctan(y / x) from -pi to pi
		{
			int pI;
			bool xSign;
			bool ySign;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			ySign = y.S;					// save sign of y because z and y may be the same memory
			if (MultiFD.ZTestF(x))
			{
				if (MultiFD.ZTestF(y))
				{
					Calc.ClearN(zI);
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					return;
				}
				if (Calc.degMode)
					SetTo1N(90, zI);	// z = 90
				else
				{
					pI = GetPiPtr();
					Calc.SetToII(pI, zI);
					RDiv1N(2, zI);		// z = pi/2
				}
				Common.nList[zI].S = ySign;
			}
			else
			{
				xSign = x.S;
				DivN(y, x, zI);
				RATanN(zI);
				if (xSign)
				{
					if (Calc.degMode)
					{
						if (ySign)
							RAdd1N(-180, zI);
						else
						{
							RAdd1N(180, zI);
						}
					}
					else
					{
						pI = GetPiPtr();
						if (ySign)
							RSubN(Common.nList[pI], zI);
						else
						{
							RAddN(Common.nList[pI], zI);
						}
					}
				}
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ATan2N

		public static void MagN(MultiF x, MultiF y, int zI)  // z = SqRt(Sq(x) + Sq(y))
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SqF(x, t);
			SqN(y, zI);
			RAddN(t, zI);
			MultiFD.DoneF(t);
			RSqRtN(zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void MeanCN(MultiF x, MultiF y, int zI)  // z = common mean of x and y
		{
			if (MultiFD.ZTestF(x) || MultiFD.ZTestF(y))
			{
				Calc.ClearN(zI); // common mean of zero and a is zero
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF a = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF s = new MultiF(Calc.fMC);
			MultiF ar = new MultiF(Calc.fMC - 2);
			MultiF gr = new MultiF(Calc.fMC - 2);
			bo = a | x;
			bo = g | y;
			a.S = false;
			g.S = false;
			for (int n = 1; true; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteLn("MeanCN: n = " + n);
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{ 
					Calc.ClearN(zI);
					break;					// exit for
				}
				bo = s | a + g;				// s = sum of a & g
				bo = g | a * g;				// g = product pf a & g
				MultiFD.RSqRtF(g);			// g = geometric mean of a & g
				bo = a | s / 2;				// a = arithmetic mean of a & g
				MultiFD.SetToF(g, gr);		// gr = g reduced in precision
				MultiFD.SetToF(a, ar);		// ar = a reduced in precision
				if (MultiFD.CompFF(ar, gr) == 0) // if a = g
				{
					Calc.SetToN(a, zI);
					break;					// exit for
				}
			}
			MultiFD.DoneF(s);
			MultiFD.DoneF(g);
			MultiFD.DoneF(a);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void ModN(MultiF x, MultiF y, int zI)  // z = Mod(x, y))
		// same as FORTRAN definition: Mod(x, y) = x - (Math.Floor(x/y) * y)
		{
			int saveFR;
			saveFR = MultiFD.fRound;
			MultiFD.fRound = 0;
			if (x.C >= (y.N + (decimal)y.C)) // not in DOS VPCalc
				Calc.ClearN(zI);
			else if ((x.C >= 0) && (y.C >= 0))
			{
				MultiF t1 = new MultiF((int)(x.N + x.C));
				MultiF t2 = new MultiF((int)(y.N + y.C));
				MultiFD.SetToF(x, t1);
				MultiFD.SetToF(y, t2);
				MultiFD.ShiftL(t1.C, t1);
				MultiFD.ShiftL(t2.C, t2);
				MultiID.RModSI(t2, t1);
				Calc.SetToN(t1, zI);
				MultiFD.DoneF(t2);
				MultiFD.DoneF(t1);
			}
			else
			{
				MultiF t = new MultiF(Calc.fMC);
				MultiFD.DivF(x, y, t);
				MultiFD.RIntF(t);
				MultiFD.RMulF(y, t);
				SubN(x, t, zI);
				MultiFD.DoneF(t);
			}
			MultiFD.fRound = saveFR;
		} // ModN

		public static void RModN(MultiF x, int zI) // z = Mod(z, x))
		{
			ModN(Common.nList[zI], x, zI);
		}

		public static void GamPN(MultiF a, MultiF x, int zI) // z = GamP(a, x)
		// GamP(a, x) is the lower regularized incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a)) // GamP(0, x > 0) = 1
			{
				if (x.S || MultiFD.ZTestF(x))
				{
					MultiID.MuWriteErr("Cannot take GamP(a, x) if a is zero and x <= zero");
					Calc.ClearN(zI);
				}
				else 
					Calc.SetToN(MultiFD.fOne, zI);
				return;
			}
			if ((a.C >= 0) && a.S)
			{
				MultiID.MuWriteErr("Cannot take GamP(a, x) if a is an integer < zero");
				Calc.ClearN(zI);
				return;
			}
			if (x.S || MultiFD.ZTestF(x)) // x <= 0.0)
			{
				MultiID.MuWriteErr("Cannot take GamP(a, x) if x <= 0.0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int tI, t2I;
			Common.InitNAdd(1, "GamPN-tI", out tI);
			Common.InitNAdd(1, "GamPN-t2I", out t2I);
			GamLN(a, x, tI);
			GamN(a, t2I);
			MultiF b = new MultiF(Calc.fMC);
			MultiFD.DivF(Common.nList[tI], Common.nList[t2I], b);	// GamP(a, x) = GamL(a, x) / Gam(a)
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiFD.DoneF(b);
			Common.LDelete(t2I);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamPN

		public static void GamQN(MultiF a, MultiF x, int zI) // z = GamQ(a, x)
		// GamQ(a, x) is the upper regularized incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a)) // GamQ(0, x > 0) = 0
			{
				if (x.S || MultiFD.ZTestF(x))
					MultiID.MuWriteErr("Cannot take GamQ(a, x) if a is zero and x <= zero");
				Calc.ClearN(zI);
				return;
			}
			if ((a.C >= 0) && a.S)
			{
				MultiID.MuWriteErr("Cannot take GamQ(a, x) if a is an integer < zero");
				Calc.ClearN(zI);
				return;
			}
			if (x.S || MultiFD.ZTestF(x)) // x <= 0.0)
			{
				MultiID.MuWriteErr("Cannot take GamQ(a, x) if x <= 0.0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int tI, t2I;
			Common.InitNAdd(1, "GamQN-tI", out tI);
			Common.InitNAdd(1, "GamQN-t2I", out t2I);
			GamUN(a, x, tI);
			GamN(a, t2I);
			MultiF b = new MultiF(Calc.fMC);
			MultiFD.DivF(Common.nList[tI], Common.nList[t2I], b);	// GamQ(a, x) = GamU(a, x) / Gam(a)
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiFD.DoneF(b);
			Common.LDelete(t2I);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamQN

		public static void GamLN(MultiF a, MultiF x, int zI) // z = GamL(a, x)
		// GamL(a, x) is the lower incomplete gamma function with parameter a and argument x
		{
			if ( ((a.C >= 0) && a.S) || MultiFD.ZTestF(a) )
			{
				MultiID.MuWriteErr("Cannot take GamL(a, x) if a is an integer <= zero");
				Calc.ClearN(zI);
				return;
			}
			if (x.S || MultiFD.ZTestF(x)) // x <= 0.0)
			{
				MultiID.MuWriteErr("Cannot take GamL(a, x) if x <= 0.0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF b = new MultiF(Calc.fMC);
			bo = b | a + 1;
			MultiF lim = new MultiF(1);
			bo = lim | 100;
			if ( (MultiFD.CompFF(b, x) < 0) ||
				 (MultiFD.CompFF(lim, x) < 0) ) // if x < a + 1 or x < 100
			{
				GamLS(a, x, b); // b = GamL(a, x) by series
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "GamLN-tI", out tI);
				GamUC(a, x, b); // b = GamU(a, x) by continued fraction
				GamNF(a, tI);	// t = Gamma(a)
				bo = b | Common.nList[tI] - b; // b = t - b = Gamma(a) - GamU(a, x)
				Common.LDelete(tI);
			}
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiFD.DoneF(b);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamLN

		public static void GamUN(MultiF a, MultiF x, int zI) // z = GamU(a, x)
		// GamU(a, x) is the upper incomplete gamma function with parameter a and argument x
		{
			if (MultiFD.ZTestF(a)) // GamQ(0, x > 0) = E1(x)
			{
				if (x.S || MultiFD.ZTestF(x))
				{
					MultiID.MuWriteErr("Cannot take GamU(a, x) if a is zero and x <= zero");
					Calc.ClearN(zI);
				}
				else
					E1N(x, zI);
				return;
			}
			if ((a.C >= 0) && a.S)
			{
				MultiID.MuWriteErr("Cannot take GamU(a, x) if a is an integer < zero");
				Calc.ClearN(zI);
				return;
			}
			if (x.S || MultiFD.ZTestF(x)) // x <= 0.0)
			{
				MultiID.MuWriteErr("Cannot take GamU(a, x) if x <= 0.0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF b = new MultiF(Calc.fMC);
			bo = b | a + 1;
			MultiF lim = new MultiF(1);
			bo = lim | 100;
			if ( (MultiFD.CompFF(b, x) < 0) ||
				 (MultiFD.CompFF(lim, x) < 0) ) // if x < a + 1 or x < 100
			{
				int tI;
				Common.InitNAdd(1, "GamLN-tI", out tI);
				GamLS(a, x, b); // b = GamL(a, x) by series
				GamNF(a, tI);	// t = Gamma(a)
				bo = b | Common.nList[tI] - b; // b = t - b = Gamma(a) - GamU(a, x)
				Common.LDelete(tI);
			}
			else
			{
				GamUC(a, x, b); // b = GamU(a, x) by continued fraction
			}
			Calc.NormNM(b);
			Calc.SetToN(b, zI);
			MultiFD.DoneF(b);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // GamUN

		public static void GamUC(MultiF a, MultiF x, MultiF z) // z = GamU(a, x)
		// evaluated by its continued fraction representation
		{
			MultiF g = new MultiF(Calc.fMC);
			MultiF gr = new MultiF(Calc.fMC - 2);
			MultiF gOld = new MultiF(Calc.fMC - 2);
			MultiF a0 = new MultiF(Calc.fMC);
			MultiF a1 = new MultiF(Calc.fMC);
			MultiF b0 = new MultiF(Calc.fMC);
			MultiF b1 = new MultiF(Calc.fMC);
			MultiF fac = new MultiF(Calc.fMC);
			MultiF anf = new MultiF(Calc.fMC);
			MultiF an = new MultiF(Calc.fMC);
			MultiF ana = new MultiF(Calc.fMC);
			MultiF t = new MultiF(Calc.fMC);
			bo = gOld | 0;
			bo = a0 | 1;
			bo = a1 | x;
			bo = b0 | 0;
			bo = b1 | 1;
			bo = fac | 1;
			bo = an | 0;
			for (int n = 1; true; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteLn("GamUC: n = " + n);
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{ 
					bo = z | 0;
					break;								// exit for
				}
				MultiFD.AddF(an, MultiFD.fOne, an);		// an = n
				MultiFD.SubF(an, a, ana);				// ana = an - a;
				MultiFD.MulF(a0, ana, a0);				// a0 = (a1 + a0 * ana) * fac
				MultiFD.AddF(a0, a1, a0);
				MultiFD.MulF(a0, fac, a0);
				MultiFD.MulF(b0, ana, b0);				// b0 = (b1 + b0 * ana) * fac
				MultiFD.AddF(b0, b1, b0);
				MultiFD.MulF(b0, fac, b0);
				MultiFD.MulF(an, fac, anf);				// anf = an * fac
				MultiFD.MulF(x, a0, t);					// a1 = x * a0 + anf * a1
				MultiFD.MulF(anf, a1, a1);
				MultiFD.AddF(t, a1, a1);
				MultiFD.MulF(x, b0, t);					// b1 | x * b0 + anf * b1
				MultiFD.MulF(anf, b1, b1);
				MultiFD.AddF(t, b1, b1);
				if (!MultiFD.ZTestF(a1))
				{
					MultiFD.DivF(MultiFD.fOne, a1, fac);	// fac = 1 / a1
					MultiFD.MulF(b1, fac, g);				// g = b1 * fac
					MultiFD.SetToF(g, gr);					// gr = g reduced in precision
					if (MultiFD.CompFF(gr, gOld) == 0)		// if gr = gOld
					{
						int tI;
						Common.InitNAdd(1, "GamUC-tI", out tI);
						XPCalLst.LnN(x, tI);						// t = ln(x)
						bo = an | a * Common.nList[tI] - x;			// an = a * ln(x) - x
						XPCalLst.ExpN(an, tI);						// t = exp(a * ln(x) - x)
						MultiFD.MulF(g, Common.nList[tI], z);		// z = g * t
						Common.LDelete(tI);
						break;
					}
					bo = gOld | gr;									// save old gr
				}
			}
			MultiFD.DoneF(t);
			MultiFD.DoneF(ana);
			MultiFD.DoneF(an);
			MultiFD.DoneF(anf);
			MultiFD.DoneF(fac);
			MultiFD.DoneF(b1);
			MultiFD.DoneF(b0);
			MultiFD.DoneF(a1);
			MultiFD.DoneF(a0);
			MultiFD.DoneF(gOld);
			MultiFD.DoneF(gr);
			MultiFD.DoneF(g);
    	} // GamUC

		public static void GamLS(MultiF a, MultiF x, MultiF z) // z = GamL(a, x)
		// evaluated by its series representation
		{
			int n;
			MultiF sum = new MultiF(Calc.fMC);
			MultiF del = new MultiF(Calc.fMC);
			MultiF aP = new MultiF(Calc.fMC);
			MultiF b = new MultiF(Calc.fMC);
			if (x.S || MultiFD.ZTestF(x)) // x <= 0.0)
			{
				if (x.S) MultiID.MuWriteErr("GamLS: x <= 0");
				bo = z | 0;
				return;
			}
			bo = aP | a;
			MultiFD.DivF(MultiFD.fOne, a, sum);		// sum = 1 / a
			bo = del | sum;							// del = sum
			for (n = 1; ; n++)
			{
				MultiID.MuInterrupt();
				if (MultiID.statReq2)
				{
					Common.SingleSpace();
					MultiID.WriteLn("GamLS: n = " + n);
					MultiID.statReq2 = false;
				}
				if (MultiID.muAbort || MultiID.muErr)
				{ 
					bo = z | 0;
					break;						// exit for
				}
				aP += 1;
				MultiFD.DivF(x, aP, b);			// b = x / aP
				MultiFD.MulF(del, b, del);		// del = del * x / aP
				bo = b | sum;					// b = previous sum
				sum += del;
				bo = del | sum - b; 
				if (MultiFD.ZTestF(del))
				{
					int tI;
					Common.InitNAdd(1, "GamLS-tI", out tI);
					XPCalLst.LnN(x, tI);						// t = ln(x)
					bo = aP | a * Common.nList[tI] - x;			// aP = a * ln(x) - x
					XPCalLst.ExpN(aP, tI);						// t = exp(a * ln(x) - x)
					MultiFD.MulF(sum, Common.nList[tI], z);		// z = sum * t
					Common.LDelete(tI);
					break;
				}
			}
			// MultiID.MuWriteErr("GamLS: a too large");
			MultiFD.DoneF(b);
			MultiFD.DoneF(aP);
			MultiFD.DoneF(del);
			MultiFD.DoneF(sum);
		} // GamLS

		public static void BinoN(MultiF x, MultiF y, int zI)
		// z = Binomial coefficient (x, y), uses the integer part of y
		{
			MultiF yL = new MultiF((int)Math.Min(y.N, y.N + y.C));
			MultiFD.IntF(y, yL);						// 1: yL = integer part of y
			if (yL.S)									// 2: (x, y) = 0 for all x if y < 0
			{
				MultiFD.DoneF(yL);
				Calc.ClearN(zI);
				return;
			}
			bool sign = false;							// set sign +
			bool xInt = false;
			MultiF xL = new MultiF(Calc.fMC);
			bo = xL | x;
			if (xL.C >= 0)								// 3: if x is an integer
			{
				xInt = true;
				if (xL.S)								// 3.1: if x < 0
				{
					bo = xL | (yL - xL - 1);			// (x, y) = (-1)^y * (y-x-1, y)
					sign = ( (yL.C == 0) && ((yL.V[0] & 1) == 1) ); // set sign true if y is odd
				}
				if (MultiFD.CompFF(xL, yL) > 0)			// 3.2: if y > x
				{
					MultiFD.DoneF(xL);
					MultiFD.DoneF(yL);
					Calc.ClearN(zI);					// (x, y) = 0
					return;
				}
				MultiF h = new MultiF(xL.N+1);
				bo = h | xL / 2;
				if (MultiFD.CompFF(h, yL) > 0)			// 3.3: y > x / 2
				{
					bo = yL | xL - yL;					// y = x - y
				}
				MultiFD.DoneF(h);
			}
			if (MultiFD.ZTestF(yL))						// 4: (x, 0) = 1 for all x
			{
				Calc.SetToN(MultiFD.fOne, zI);
				Common.nList[zI].S = sign;
				MultiFD.DoneF(xL);
				MultiFD.DoneF(yL);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF v = new MultiF(Calc.fMC);
			bo = v | Int32.MaxValue;
			if (MultiFD.CompFF(v, yL) > 0)				// 5: y > 2^31 -1
			{
				Calc.ClearN(zI);						// (x, y) = 0 with error message
				MultiID.MuWriteErr("BinoN: numbers too large for binomial coefficient");
			}
			else
			{
				MultiF nf = new MultiF(Calc.fMC);		// 6: (x, y) = x * (x-1) * (x-2) * ... * (x-y+1)
				MultiF df = new MultiF(Calc.fMC);		//			/ 1 * 2 * ... * n
				bo = nf | xL;
				bo = df | 1;
				bo = v | nf;
				double yd;
				MultiFD.GetDbF(out yd, yL);
				int yi = (int)yd;
				for (int i = 1; i < yi; i++)
				{
						MultiID.MuInterrupt();
						if (MultiID.muAbort)
						{
							MultiFD.ClearF(v);
							sign = false;
							break;
						}
						if (MultiID.statReq2)
						{
							MultiID.WriteLn("BinoN: Done " + i + " of " + yi);
							RunForm.FocusBothUs();
							MultiID.statReq2 = false;
						}
						MultiFD.AddF(nf, MultiFD.fMOne, nf);	// nf = nf - 1
						MultiFD.AddF(df, MultiFD.fOne, df);		// df = df + 1
						MultiFD.MulF(v, nf, v);					// v = (v * nf) / df
						MultiFD.DivF(v, df, v);
				}
				Calc.SetToN(v, zI);
				if (xInt) Common.nList[zI].S = sign;
				MultiFD.DoneF(df);
				MultiFD.DoneF(nf);
			}
			MultiFD.DoneF(v);
			MultiFD.DoneF(xL);
			MultiFD.DoneF(yL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BinoN

		public static void BetaCN(MultiF x, MultiF y, int zI)
		// z = the complete beta function B(x, y) = gamma(x) * gamma(y) / gamma(x+y)
		// also equal to 1 / (x * Bino(x+y-1, y-1)
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF xL = new MultiF(Calc.fMC);
			bo = xL | x + y;
			bool gxInf = ((x.C >= 0) && x.S) || MultiFD.ZTestF(x); // gamma(x) is infinite
			bool gyInf = ((y.C >= 0) && y.S) || MultiFD.ZTestF(y); // gamma(y) is infinite
			bool gxPyInf = ((xL.C >= 0) && xL.S) || MultiFD.ZTestF(xL); // gamma(x+y) is infinite
			if (!gxInf && !gyInf && gxPyInf)
			{
				Calc.ClearN(zI);					// finite / infinity = 0
			}
			else
			{
				if ( ((gxInf || gyInf) && !gxPyInf) || (gxInf && gyInf) )	// infinity / finite = infinity
				{															// infinity * infinity / v = infinity
					MultiID.MuWriteErr("BetcC(x' y) error: BetaC is infinite");
					Calc.ClearN(zI);
				}
				else
				{
					if (gxPyInf)											// finite * infinity / infinity = ?
					{
						if (gxInf)
						{
							MultiF t = x; // interchange pointers to x and y so gamma(y) is infinity
							x = y;
							y = t;
						}
						bool sign = ( (y.C == 0) && ((y.V[0] & 1) == 1) ) ^ // set sign true if y is odd
							( (xL.C == 0) && ((xL.V[0] & 1) == 1) );		// exclusive-or x+y is odd
						bo = xL | 1 - xL;									// xL = 1 - (x+y)
						MultiF yL2 = new MultiF(Calc.fMC);
						bo = yL2 | 1 - y;									// yL2 = 1 - y
						int t2I;
						Common.InitNAdd(1, "BetaCN-t2I", out t2I);			// t2 = temp on the list
						GamN(xL, t2I);
						bo = xL | Common.nList[t2I];						// xL = gamma(1 - (x+y))
						GamN(x, t2I);
						xL *= Common.nList[t2I];							// xL = gamma(x) * gamma(1 - (x+y))
						GamN(yL2, t2I);
						xL.S = sign;
						DivN(xL, Common.nList[t2I], zI);                    // B = xL / gamma(1 - y)
						Common.LDelete(t2I);
						MultiFD.DoneF(yL2);
					}
					else
					{
						xL -= 1;
						MultiF yL = new MultiF(Calc.fMC);
						bo = yL | y - 1;
						int tI;
						Common.InitNAdd(1, "BetaCN-tI", out tI);			// t = temp on the list
						BinoGN(xL, yL, tI);
						bo = xL | x * Common.nList[tI];
						InvN(xL, zI);
						Common.LDelete(tI);
						MultiFD.DoneF(yL);
					}
				}
			}
			MultiFD.DoneF(xL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BetaCN

		public static void BinoGN(MultiF x, MultiF y, int zI)
		// z = Binomial coefficient (x, y) using the gamma function
		{
			if ( (MultiFD.ZTestF(y)) || (MultiFD.CompFF(x, y) == 0) )
			{
				Calc.SetToN(MultiFD.fOne, zI);	// (x, 0) = (x, x) = 1 for all x
				return;
			}
			if (MultiFD.CompFF(MultiFD.fOne, y) == 0)
			{
				Calc.SetToN(x, zI);				// (x, 1) = x for all x
				return;
			}
			bool sign = false;
			MultiF xL = new MultiF(Calc.fMC);
			bo = xL | x;
			if (y.C >= 0)						// if y is an integer
			{
				if (y.S)						// if y is a negative integer, (x, y) = 0
				{
					MultiFD.DoneF(xL);
					Calc.ClearN(zI);
					return;
				}
				if (xL.C >= 0)					// if x is an integer
				{
					if (xL.S)										// 3.1: if x < 0
					{
						bo = xL | (y - xL - 1);						// (x, y) = (-1)^y * (y-x-1, y)
						sign = ( (y.C == 0) && ((y.V[0] & 1) == 1) ); // set sign true if y is odd
						if (MultiFD.CompFF(xL, y) == 0)
						{
							MultiFD.DoneF(xL);
							Calc.SetToN(MultiFD.fOne, zI);			// (a, a) = 1 for all a
							Common.nList[zI].S = sign;
							return;
						}
					}
					if (MultiFD.CompFF(xL, y) > 0)					// 3.2: if y > x
					{
						MultiFD.DoneF(xL);
						Calc.ClearN(zI);							// (x, y) = 0
						return;
					}
				}
				MultiF lim = new MultiF(1);
				bo = lim | limFac + 1;
				if (MultiFD.CompFF(lim, y) <= 0)	    // if y is a small integers
				{
					MultiFD.DoneF(lim);
					BinoN(xL, y, zI);
					MultiFD.DoneF(xL);
					Common.nList[zI].S = Common.nList[zI].S ^ sign; // ^ is xor
					return;
				}
				MultiFD.DoneF(lim);
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF yL = new MultiF(Calc.fMC);
			bo = yL | xL - y;			// yL = x - y
			if ((yL.C >= 0) && yL.S)	// if gamma of x - y + 1 would be infinite
			{
				Calc.ClearN(zI);		// divide by infinity gives zero
			}
			else
			{
				if (MultiFD.CompFF(MultiFD.fOne, yL) == 0)
				{
					Calc.SetToN(xL, zI);	// (a, a-1) = a for all a
				}
				else
				{
					bo = yL | yL + 1;
					bo = xL | xL + 1;
					int tI;
					Common.InitNAdd(1, "BinoGN-tI", out tI); // t = temp on the list
					GamNF(xL, tI);
					if (MultiID.muErr || MultiID.muAbort)
					{
						Calc.ClearN(zI);
					}
					else
					{
						int t2I;
						Common.InitNAdd(1, "BinoGN-t2I", out t2I); // t2 = temp on the list
						GamNF(yL, t2I);
						MultiF v = new MultiF(Calc.fMC);
						bo = v | Common.nList[tI] / Common.nList[t2I];
						bo = yL | y + 1;
						GamNF(yL, t2I);
						v /= Common.nList[t2I];
						Calc.SetToN(v, zI);
						Common.nList[zI].S = Common.nList[zI].S ^ sign; // ^ is xor
						MultiFD.DoneF(v);
						Common.LDelete(t2I);
					}
					Common.LDelete(tI);
				}
			}
			MultiFD.DoneF(yL);
			MultiFD.DoneF(xL);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // BinoGN

		public static void GCDN(MultiF x, MultiF y, int zI) // z = GCD(x, y))
		{
			int saveFR;
			saveFR = MultiFD.fRound;
			MultiFD.fRound = 0;
			MultiF t1 = new MultiF((int)(x.N + x.C));
			MultiF t2 = new MultiF((int)(y.N + y.C));
			MultiFD.SetToF(x, t1);
			MultiFD.SetToF(y, t2);
			MultiFD.ShiftL(t1.C, t1);
			MultiFD.ShiftL(t2.C, t2);
			MultiID.RGCDSI(t2, t1);
			Calc.NormNM(t1);
			Calc.SetToN(t1, zI);
			MultiFD.DoneF(t2);
			MultiFD.DoneF(t1);
			MultiFD.fRound = saveFR;
		}

		public static void LCMN(MultiF x, MultiF y, int zI) // z = LCM(x, y))
		{
			int saveFR;
			saveFR = MultiFD.fRound;
			MultiFD.fRound = 0;
			MultiF t1 = new MultiF((int)(x.N + x.C));
			MultiF t2 = new MultiF((int)(y.N + y.C));
			MultiFD.SetToF(x, t1);
			MultiFD.SetToF(y, t2);
			MultiFD.ShiftL(t1.C, t1);
			MultiFD.ShiftL(t2.C, t2);
			MultiID.RLCMSI(t2, t1);
			Calc.NormNM(t1);
			Calc.SetToN(t1, zI);
			MultiFD.DoneF(t2);
			MultiFD.DoneF(t1);
			MultiFD.fRound = saveFR;
		}

		public static void ASinN(MultiF x, int zI)  // z = ASin(x)
		// ASin(S) = ATan2(S, SqRt(1 - Sq(S)), but if Sq(S) > 1 then Error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			if (x.C < (-Int32.MaxValue / 3 - Int16.MaxValue))
			{
				Calc.SetToN(x, zI);
				if (Calc.degMode) RToDegN(zI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.NormN(zI);
				return;
			}
			bool saveRep = MultiID.muErrRep;
			MultiID.muErrRep = false;
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SqF(x, t);
			t.S = true;
			MultiFD.RAdd1F(1, t); // 1 - S^2
			MultiID.muErrRep = saveRep;
			if (MultiID.muErr || t.S)
			{
				MultiID.MuWriteErr("ASin(x) error: |x| > 1.0");
				Calc.ClearN(zI);
			}
			else
			{
				MultiFD.RSqRtF(t);
				ATan2N(x, t, zI);
			}
			MultiFD.DoneF(t);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ASinN

		public static void RASinN(int zI) // z = ASin(z)
		{
			ASinN(Common.nList[zI], zI);
		}

		public static void ACosN(MultiF x, int zI)  // z = ACos(x)
		// ACos(C) = ATan2(SqRt(1 - Sq(C), C), but if Sq(C) > 1 then Error
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			bool saveRep = MultiID.muErrRep;
			MultiID.muErrRep = false;
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SqF(x, t);
			t.S = true;
			MultiFD.RAdd1F(1, t); // 1 - C^2
			MultiID.muErrRep = saveRep;
			if (MultiID.muErr || t.S)
			{
				MultiID.MuWriteErr("ACos(x) error: |x| > 1.0");
				Calc.ClearN(zI);
			}
			else
			{
				MultiFD.RSqRtF(t);
				ATan2N(t, x, zI);
			}
			MultiFD.DoneF(t);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ACosN

		public static void RACosN(int zI) // z = ACos(z)
		{
			ACosN(Common.nList[zI], zI);
		}

		public static void SExpLN(MultiF x, bool bL1, int zI)
		// z= (e^x) - 1, small x, (- 1 only if bL1 is true)
		// Exp(x) - 1 = x + x^2/2! + x^3/3! + ...
		{
			int saveFMC;
			int saveMMax;
			int currFMC;
			bool fini;
			int j;
			if ( bL1 && (x.C < (-Int32.MaxValue / 2 - Int16.MaxValue)) )
			{
				Calc.SetToN(x, zI);
				return;
			}
			saveFMC = Calc.fMC;
			saveMMax = MultiID.mMax;
			currFMC = Calc.fMC;
			MultiF nn = new MultiF(2);
			MultiF term = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			MultiFD.SetTo1F(1, nn);				// nn	= 1
			MultiFD.SetToF(x, term);			// term	= x
			MultiFD.SetToF(term, g);			// g	= term
			MultiFD.SetToF(term, sum);			// sum	= term
			if (!bL1) MultiFD.RAdd1F(1, sum);
			fini = false;
			MultiID.nDone = 1;
			while (!fini)
			{
				MultiFD.RAdd1F(1, nn);			// nn++
				MultiFD.RMulF(g, term);			// term = (term * g) / nn
				MultiFD.RDivF(nn, term);
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				long iL = Calc.fMC - sum.N - sum.C + term.N + term.C;
				if (iL > Int32.MaxValue) iL = Int32.MaxValue;
				if (iL < Int32.MinValue) iL = Int32.MinValue;
				MultiID.nToGo = (int)iL;
				MultiID.nDone++;
				if ((MultiID.nToGo < 0) || (MultiFD.ZTestF(term)) || (MultiID.muErr) || (MultiID.muAbort))
					fini = true;
				else
				{
					MultiFD.RAddF(term, sum);	// sum  = sum + term
					Calc.fMC = MultiID.nToGo + 1;
					if (Calc.fMC < currFMC)
					{
						currFMC = Calc.fMC;
						j = term.N - Calc.fMC;
						if (j > 0)
							MultiFD.ShiftR(j, term);
						j = g.N - Calc.fMC;
						if (j > 0)
						{
							MultiFD.ShiftR(j, g);
						}
					}
					else
					{
						Calc.fMC = currFMC;
					}
				}
			}
			MultiID.nToGo = 0;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToN(sum, zI);
			MultiFD.DoneF(sum);
			MultiFD.DoneF(g);
			MultiFD.DoneF(term);
			MultiFD.DoneF(nn);
		} // SExpLN

		public static void MExpLN(MultiF x, bool bL1, int zI)
		// z = (e^x) - 1, medium small x, (- 1 only if bL1 is true)
		// Exp(x) = Exp(x / 2^N) Squared N times
		// since (1+a)^2 = 1 + 2*a + a*a = 1 + a * (2 + a)
		// a0 = (e^x) - 1
		// a1 = a0 * (2 + a0)
		// a2 = a0 * (2 + a1) etc.
		{
			int nn;
			double d;
			int i;
			int j;
			int k;
//			d = 3.418 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.90631455143;	// PolyFit
			d = 3.418 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.9;				// PolyFit
			Common.LFind("MExpLNN", out nn);
			if (nn != 0) MultiFD.GetDbF(out d, Common.nList[nn]);
			j = (int)Math.Round(d);
			if (MultiID.diagOn)
				MultiID.WriteLn("MExpL: NN = " + d.ToString("0.000"));
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SetToF(x, t);
			MultiF e = new MultiF(2);
			MultiFD.SetTo1F(3, e);
			for (i = 1; i <= j; i++)			// e = e / 2, nn times
				MultiFD.RDiv1F(2, e);
			MultiID.nDone = 0;
			k = j;
			MultiID.nToGo = k;
			j = 0;
			Calc.ti0 = MultiID.TimerF();
			while (MultiFD.CompFF(e, t) > 0)	// t = t / 2, until t < e
			{
				MultiFD.RDiv1F(2, t);
				j++;
				if (MultiID.muErr || MultiID.muAbort) MultiFD.SetTo1F(0, t);
				MultiID.nDone++;
				MultiID.nToGo = k - MultiID.nDone;
			}
			Calc.ti1 = MultiID.TimerF() - Calc.ti0;;
			if (MultiID.diagOn)
				MultiID.WriteLn("MExpL: NN = " + j);
			MultiFD.DoneF(e);
			Calc.ti0 = MultiID.TimerF();
			SExpLN(t, true, zI);				// A0 = ExpL(t)
			d = MultiID.TimerF();
			Calc.ti2 = d - Calc.ti0;
			Calc.ti0 = d;
			for (i = 1; i <= j; i++)			// A1 = A0 * (2 + A0), nn times
			{
				MultiFD.Add1F(Common.nList[zI], 2, t);
				RMulN(t, zI);
				MultiID.nToGo = j - i;
				MultiID.nDone = i;
			}
			if (MultiID.diagOn) ReportBest(j);
			if (!bL1) RAdd1N(1, zI);
			MultiFD.DoneF(t);
		} // MExpLN

		public static void ExpLN(MultiF x, bool bL1, int zI)
		// z = (e^x) - 1, accurate for small x, (- 1 only if bL1 is true)
		// n = Math.Floor(x / Ln(10))
		// g = Frac(x / Ln(10)) * Ln(10)
		// answer = Exp(g) * 10^n
		{
			int ln10 = 0;
			int i;
			int j;
			int d1;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF t = new MultiF(Calc.fMC);
			int ii = 1 + (25 / MultiID.muDMax);
			MultiF nn = new MultiF(ii);
			if (x.N + x.C > 0) // if |x| >= 1
			{
				if ( ((x.N + x.C) > 1) || (x.V[-x.C] > 1) ) // if |x| >= 2
				{
					ln10 = GetLn10Ptr();
					MultiFD.DivF(x, Common.nList[ln10], t);
					MultiFD.IntF(t, nn);
				}
			}
			if (MultiFD.ZTestF(nn))
				MExpLN(x, bL1, zI);
			else
			{
				if (ln10 == 0) ln10 = GetLn10Ptr();
				FracN(t, zI);
				RMulN(Common.nList[ln10], zI);
				MExpLN(Common.nList[zI], false, zI);
				decimal nDe = nn.N;
				nDe += nn.C;
				if (nDe > ii)
					nDe = 8 * (decimal)Int64.MaxValue;
				else
				{
					decimal nD;
					MultiFD.GetDeF(out nD, nn);
					nDe = nD;
				}
				if (nn.S && (nDe > 0))
					nDe = -nDe;
				decimal nID = decimal.Floor(nDe / MultiID.muDMax);
				j = (int)(nDe - nID * MultiID.muDMax);
				if (j < 0)
				{
					j += MultiID.muDMax;
					nID--;
				}
				if ((j > 0) && (j < MultiID.muDMax))
				{
					d1 = 10;
					for (i = 2; i <= j; i++)
						d1 *= 10;
					RMul1N(d1, zI);
				}
				nDe = Common.nList[zI].C;
				nDe += nID;
				if (nDe > Int64.MaxValue)
				{
					Calc.ClearN(zI);
					MultiID.MuWriteErr("Exp overflow, continuing...");
				}
				else if (nDe < Int64.MinValue)
				{
					Calc.ClearN(zI);
					if (bL1)
						Common.nList[zI].V[0] = 1;
					else
					{
						MultiID.MuWriteErr("Exp underflow, continuing...");
						MultiID.muUnderflow = true;
					}
					Common.ChSN(zI);
				}
				else
				{
					Common.nList[zI].C = (long)(Common.nList[zI].C + nID);
					if (bL1)
					{
						RAdd1N(-1, zI);
					}
				}
			}
			MultiFD.DoneF(nn);
			MultiFD.DoneF(t);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ExpLN

		public static void RExpLN(int zI) // z = (e^z) - 1, accurate for small x
		{
			ExpLN(Common.nList[zI], true, zI);
		}

		public static void ExpN(MultiF x, int zI)  // z = e^x
		{
			ExpLN(x, false, zI);
		}

		public static void RExpN(int zI) // z = e^z
		{
			ExpN(Common.nList[zI], zI);
		}

		public static void SLnLN(MultiF y, bool bL1, int zI)
		// Common.nList[zI] = Ln(y + 1), small y
		// bL1 is true if y is input, false if x = 1 + y is input
		// Ln(x) = Ln(1+y) = Ln((1+z)/(1-z)) = 2 * (z + z^3/3 + z^5/5)
		// x = 1 + y = (1 + z) / (1 - z)
		// y = x - 1 = 2 * z / (1 - z)
		// z = (x - 1) / (x + 1) = y / (2 + z)
		{
			int saveFMC;
			int saveMMax;
			int currFMC;
			bool fini;
			int j;
			if ( bL1 && (y.C < (-Int32.MaxValue / 2 - Int16.MaxValue)) )
			{
				Calc.SetToN(y, zI);
				return;
			}
			saveFMC = Calc.fMC;
			saveMMax = MultiID.mMax;
			currFMC = Calc.fMC;
			MultiF nn = new MultiF(2);
			MultiF term = new MultiF(Calc.fMC);
			MultiF bTerm = new MultiF(Calc.fMC);
			MultiF g = new MultiF(Calc.fMC);
			MultiF sum = new MultiF(Calc.fMC);
			if (!bL1)
			{
				MultiFD.Add1F(y, -1, g);			// x input So g is y
				MultiFD.Add1F(g, 2, term);			// term = z = y / (y + 2)
				MultiFD.DivF(g, term, term);
			}
			else
			{
				MultiFD.Add1F(y, 2, term);			// term = z = y / (y + 2)
				MultiFD.DivF(y, term, term);
			}
			MultiFD.AddF(term, term, bTerm);		// bTerm = 2 * z
			MultiFD.SetTo1F(1, nn);					// nn	= 1
			MultiFD.SqF(term, g);					// g	= Sq(z)
			MultiFD.SetToF(bTerm, sum);				// sum	= bTerm
			fini = false;
			MultiID.nDone = 1;
			while (!fini)
			{
				MultiFD.RMulF(g, bTerm);			// bTerm = bTerm * g
				MultiFD.RAdd1F(2, nn);				// nn	= nn + 2
				MultiFD.DivF(bTerm, nn, term);		// term = bTerm / nn
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				long iL = Calc.fMC - sum.N - sum.C + term.N + term.C;
				if (iL > Int32.MaxValue) iL = Int32.MaxValue;
				if (iL < Int32.MinValue) iL = Int32.MinValue;
				MultiID.nToGo = (int)iL;
				MultiID.nDone++;
				if ((MultiID.nToGo < 0) || (MultiFD.ZTestF(term)) || (MultiID.muErr) || (MultiID.muAbort))
					fini = true;
				else
				{
					MultiFD.RAddF(term, sum);		// sum  = sum + term
					Calc.fMC = MultiID.nToGo + 1;
					if (Calc.fMC < currFMC)
					{
						currFMC = Calc.fMC;
						j = term.N - Calc.fMC;
						if (j > 0)
							MultiFD.ShiftR(j, term);
						j = g.N - Calc.fMC;
						if (j > 0)
						{
							MultiFD.ShiftR(j, g);
						}
					}
					else
					{
						Calc.fMC = currFMC;
					}
				}
			}
			MultiID.nToGo = 0;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.SetToN(sum, zI);
			MultiFD.DoneF(sum);
			MultiFD.DoneF(g);
			MultiFD.DoneF(bTerm);
			MultiFD.DoneF(term);
			MultiFD.DoneF(nn);
		} // SLnLN

		public static void MLnLN(MultiF y, bool bL1, int zI)
		// z = Ln(y + 1), medium small y
		// Ln(x) = Ln(y + 1) = Ln(SqRt(x) n times) * 2^n
		// SqRt(y + 1) - 1 = y / (1 + SqRt(1 + y)
		{
			int nn;
			double d;
			bool sign;
			int i;
			int j;
			int k;
//			d = 0.6795 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.56415806582;	// PolyFit
			d = 0.6795 + Math.Sqrt(Calc.fMC * MultiID.muDMax) * 0.24;			// PolyFit
			Common.LFind("MLnLNN", out nn);
			if (nn != 0) MultiFD.GetDbF(out d, Common.nList[nn]);
			j = (int)Math.Round(d);
			if (MultiID.diagOn)
				MultiID.WriteLn("MLnL: NN = " + d.ToString("0.000"));
			MultiF e = new MultiF(2);
			MultiF f = new MultiF(2);
			MultiFD.SetTo1F(9, e);
			for (i = 1; i <= j; i++)				// e = e / (1 + SqRt(1 + e)), nn times
			{
				MultiFD.Add1F(e, 1, f);
				MultiFD.RSqRtF(f);
				MultiFD.RAdd1F(1, f);
				MultiFD.RDivF(f, e);
			}
			MultiFD.DoneF(f);
			MultiF t = new MultiF(Calc.fMC);
			MultiFD.SetToF(y, t);
			if (!bL1)
				MultiFD.Add1F(y, -1, t);			// x input So t is x - 1
			sign = t.S;
			t.S = false;
			MultiID.nDone = 0;
			k = j;
			MultiID.nToGo = k;
			j = 0;
			Calc.ti0 = MultiID.TimerF();
			while (MultiFD.CompFF(e, t) > 0)		// t = t / (1 + SqRt(1 + t)), until t < e
			{
				t.S = sign;
				Add1N(t, 1, zI);
				RSqRtN(zI);
				RAdd1N(1, zI);
				MultiFD.RDivF(Common.nList[zI], t);
				j++;
				if (MultiID.muErr || MultiID.muAbort) MultiFD.SetTo1F(0, t);
				t.S = false;
				MultiID.nDone++;
				MultiID.nToGo = k - MultiID.nDone;
			}
			Calc.ti1 = MultiID.TimerF() - Calc.ti0;
			if (!MultiFD.ZTestF(t)) t.S = sign;
			if (MultiID.diagOn)
				MultiID.WriteLn("MLnL: NN = " + j);
			Calc.ti0 = MultiID.TimerF();
			SLnLN(t, true, zI);						// A0 = LnL(t)
			d = MultiID.TimerF();
			Calc.ti2 = d - Calc.ti0;
			Calc.ti0 = d;
			for (i = 1; i <= j; i++)				// A1 = A0 * 2, nn times
			{
				RMul1N(2, zI);
				MultiID.nToGo = j - i;
				MultiID.nDone = i;
			}
			if (MultiID.diagOn) ReportBest(j);
			MultiFD.DoneF(t);
			MultiFD.DoneF(e);
		} // MLnLN

		public static void LnLN(MultiF y, bool bL1, int zI)
		// z = Ln(y + 1), accurate for small y
		// x = y + 1 = Integer * MultiID.baseM^C
		// Multiply by 10 until MSD of i >= MultiID.baseM / 10, save count
		// change C of x to -N so x < 1, save original C+N
		// answer = Ln(x) + Ln(10) * NT, NT = MultiID.muDMax * C+N - count
		{
			int tI;
			decimal d;
			decimal nTen;
			int ln10;
			if (bL1)
			{
				if ( (-y.C > y.N) || ((-y.C == y.N) &&
					(y.V[y.N-1] < MultiID.baseM / 10)) )
				{
					MLnLN(y, true, zI);		// y < 0.1
					return;
				}
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "LnLN-tI", out tI);
			Calc.SetToN(y, tI);
			if (bL1) RAdd1N(1, tI);
			if (Common.ZTestN(tI))
			{
				MultiID.MuWriteErr("LnL: Cannot take Ln of zero, continuing...");
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.ClearN(zI);
				return;
			}
			if (Common.nList[tI].S)
			{
				MultiID.MuWriteErr("LnL: Cannot take Ln of X < 0, continuing...");
				Common.LDelete(tI);
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Calc.ClearN(zI);
				return;
			}
			if (Common.nList[tI].V[Common.nList[tI].N-1] == 1)
			{
				nTen = Common.nList[tI].C;
				nTen += Common.nList[tI].N - 1;		//  HJS 2004-07-04
				nTen *= MultiID.muDMax;
				Common.nList[tI].C = 1 - Common.nList[tI].N;
			}
			else
			{
				nTen = 0;
				while ((Common.nList[tI].V[Common.nList[tI].N-1] < MultiID.baseM / 10) && !MultiID.muErr && !MultiID.muAbort)
				{
					RMul1N(10, tI);
					nTen--;
				}
				d = Common.nList[tI].C;
				nTen = nTen + (d + Common.nList[tI].N) * MultiID.muDMax;
				Common.nList[tI].C = -Common.nList[tI].N;
			}
			MLnLN(Common.nList[tI], false, zI);
			if (nTen != 0)
			{
				ln10 = GetLn10Ptr();
				MultiF fTen = new MultiF(5);
				MultiFD.SetToDeF(nTen, fTen);
				MulN(Common.nList[ln10], fTen, tI);
				MultiFD.DoneF(fTen);
				RAddN(Common.nList[tI], zI);
			}
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // LnLN

		public static void RLnLN(int zI) // z = Ln(z + 1), accurate for small z
		{
			LnLN(Common.nList[zI], true, zI);
		}

		public static void LnN(MultiF x, int zI)	// z = Ln(x)
		{
			LnLN(x, false, zI);
		}

		public static void RLnN(int zI) // z = Ln(z)
		{
			LnN(Common.nList[zI], zI);
		}

		public static void LogN(MultiF x, int zI)	// z = Log(x)
		// Log(x) = Ln(x) / Ln(10), Log base 10 of x
		{
			int ln10;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			LnN(x, zI);
			ln10 = GetLn10Ptr();
			RDivN(Common.nList[ln10], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RLogN(int zI) // z = Log(z)
		{
			LogN(Common.nList[zI], zI);
		}

		public static void SinHN(MultiF x, int zI)  // z = SinH(x)
		// for (x >= 0.1 SinH(x) = (y - 1/y) / 2, where y = Exp(x)
		// for (x <  0.1 SinH(x) = y / (2 * SqRt(y + 1)), where y = Exp(2 * x) - 1
		// SinH(-x) = -SinH(x)
		{
			int tI;
			bool sign;
			Calc.SetToN(x, zI);
			if ( Common.ZTestN(zI) || (Common.nList[zI].C < (-Int32.MaxValue / 3 - Int16.MaxValue)) )
				return;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "SinHN-tI", out tI);
			sign = Common.nList[zI].S;
			Common.nList[zI].S = false;
			if ( (-Common.nList[zI].C > Common.nList[zI].N) ||
				((-Common.nList[zI].C == Common.nList[zI].N) &&
				(Common.nList[zI].V[Common.nList[zI].N-1] < (MultiID.baseM / 10))) )
			{
				RMul1N(2, zI);
				RExpLN(zI);
				Add1N(Common.nList[zI], 1, tI);
				RSqRtN(tI);
				RMul1N(2, tI);
				RDivN(Common.nList[tI], zI);
			}
			else
			{
				RExpN(zI);
				if (!Common.ZTestN(zI)) InvN(Common.nList[zI], tI);
				RSubN(Common.nList[tI], zI);
				RDiv1N(2, zI);
			}
			Common.LDelete(tI);
			Common.nList[zI].S = sign;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // SinHN

		public static void RSinHN(int zI) // z = SinH(z)
		{
			SinHN(Common.nList[zI], zI);
		}

		public static void EiN(MultiF x, int zI)  // z = entire exponential integral of x
		// Ei(x) = -Ei(-x), only use this for x << -1 (large negative x)
		{
			MultiF s = new MultiF(Calc.fMC);
			MultiF k = new MultiF(Calc.fMC);
			MultiF f = new MultiF(Calc.fMC);
			MultiF d = new MultiF(Calc.fMC);
			MultiF t = new MultiF(Calc.fMC);
			MultiF p = new MultiF(Calc.fMC);
			bo = k | 0;					// k = index
			bo = f | 1;					// f = k!
			bo = d | x;					// d = x ^ (k+1)
			bo = s | 0;					// s = sum
			for (int ki = 0; true; ki++)    // sum of (x)^k / (k*k!)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				bo = t | f / d;			// term = k! / d
				bo = p | s;				// p = previous sum
				s += t;					// sum = sum + t
				bo = p | s - p;			// p = delta in sum
				if (MultiFD.ZTestF(p))
				{
					break;
				}
				k += 1;					// update index
				f *= k;					// f = k! 
				d *= x;					// d = x ^ (k+1)
			}
			MultiFD.DoneF(p);
			MultiFD.DoneF(t);
			MultiFD.DoneF(d);
			MultiFD.DoneF(f);
			MultiFD.DoneF(k);
			if (MultiID.muAbort || MultiID.muErr)
			{
				Calc.ClearN(zI);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "EiN-tI", out tI);
				XPCalLst.ExpN(x, tI);				// tI = exp(x)
				s *= Common.nList[tI];				// times sum
				Calc.SetToN(s, zI);					// final value is s
				Common.LDelete(tI);
			}
			MultiFD.DoneF(s);
		} // EiN

		public static void E1N(MultiF x, int zI)  // z = exponential integral 1 of x
		// E1(x) = integral{x to inf}(e^-t / t)dt = -Ei(-x)
		{
			int more = 0;
			long lost = 0;
			if (MultiFD.ZTestF(x))
			{
				MultiID.MuWriteErr("E1(x) error: x = 0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF mx = new MultiF(x.N);			// mx = -x
			bo = mx | x;
			mx.S = !x.S;
			if (mx.S && (MultiFD.CompFF(MultiFD.fOne, x) > 0)) // if x > 1, subject to large relative error
			{
				MultiF lim = new MultiF(2);			// limit for using Ei(x)
				bo = lim | (int)(10 + Calc.fMC * Math.Log(MultiID.baseM));
				if (MultiFD.CompFF(lim, x) >= 0)		// if x >= limit
				{
					EiN(mx, zI);
					Common.nList[zI].S = false;			// E1 = -Ei(-x) (E1 > 0 for x > 0)
					MultiFD.DoneF(lim);
					MultiFD.DoneF(mx);
					MultiFD.expanded = saveExp;
					Calc.fMC = saveFMC;
					MultiID.mMax = saveMMax;
					Calc.NormN(zI);
					return;
				}
				double db;
				MultiFD.GetDbF(out db, x);
				more = 1 + (int)(db / 9.22);	// Need more precision
				Calc.fMC += more;
				Calc.CheckNMax();
				if (MultiID.diagOn) MultiID.WriteLn("E1: added " + more * MultiID.muDMax + " more digits");
			}
			MultiF s = new MultiF(Calc.fMC);
			MultiF k = new MultiF(Calc.fMC);
			MultiF n = new MultiF(Calc.fMC);
			MultiF f = new MultiF(Calc.fMC);
			MultiF d = new MultiF(Calc.fMC);
			MultiF t = new MultiF(Calc.fMC);
			MultiF p = new MultiF(Calc.fMC);
			bo = k | 1;					// k = index
			bo = n | mx;				// n = numerator
			bo = f | 1;					// f = factorial factor
			bo = s | mx;				// s = sum
			for (int ki = 2; true; ki++)    // sum of (mx)^k / (k*k!)
			{
				MultiID.MuInterrupt();
				if (MultiID.muAbort || MultiID.muErr) break;
				k += 1;					// update index
				n *= mx;				// x^k
				f *= k;					// f = k! 
				bo = d | f * k;			// d = k * k!
				bo = t | n / d;			// t = n / d
				bo = p | s;				// p = previous sum
				s += t;					// sum = sum + t
				if ((s.N + s.C) > lost) lost = s.N + s.C;
				bo = d | s - p;			// d = delta in sum
				if (MultiFD.ZTestF(d))
				{
					break;
				}
			}
			MultiFD.DoneF(p);
			MultiFD.DoneF(t);
			MultiFD.DoneF(d);
			MultiFD.DoneF(f);
			MultiFD.DoneF(n);
			MultiFD.DoneF(k);
			if (MultiID.muAbort || MultiID.muErr)
			{
				Calc.ClearN(zI);
			}
			else
			{
				int tI;
				Common.InitNAdd(1, "E1N-tI", out tI);
				mx.S = false;						// x = |x|
				XPCalLst.LnN(mx, tI);				// tI = ln(|x|)
				lost += s.N;
				s += Common.nList[tI];				// and add to sum
				XPCalLst.PsiN(MultiFD.fOne, tI);	// tI = -Euler's constant = -0.5772...
				s -= Common.nList[tI];				// and subtract fromto sum
				lost -= s.N;
				if (MultiID.diagOn) MultiID.WriteLn("E1: lost " + lost * MultiID.muDMax + " digits");
				if (!MultiFD.ZTestF(s)) s.S = !s.S;	// E1 = -Ei(-x) (prevent -0)
				Calc.SetToN(s, zI);					// final value is s
				Common.LDelete(tI);
			}
			MultiFD.DoneF(s);
			MultiFD.DoneF(mx);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // E1N

		public static void DilogN(MultiF x, int zI)  // z = Dilog(x)
		// Dilog(x) = x * Lerch(x, 2, 1)
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			bool xEqOne = (MultiFD.CompFF(x, MultiFD.fOne) == 0); // does x equals one?
			bool xEqMOne = (MultiFD.CompFF(x, MultiFD.fMOne) == 0); // does x equals minus one?
			if (xEqOne || xEqMOne)
			{
				int pI = GetPiPtr();
				MultiF t = new MultiF(Calc.fMC);			// t = temp
				bo = t | Common.nList[pI] * Common.nList[pI];
				if (xEqOne)
					t /= 6;									// Dilog(1) = Pi^2 / 6
				else
					t /= -12;								// Dilog(1) = -Pi^2 / 12
				Calc.SetToN(t, zI);
				MultiFD.DoneF(t);
			}
			else
			{
				MultiF saveX = new MultiF(x.N);
				bo = saveX | x;
				LerchN(x, MultiFD.fTwo, MultiFD.fOne, zI);
				XPCalLst.MulN(saveX, Common.nList[zI], zI);
				MultiFD.DoneF(saveX);
			}
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // DilogN

		public static void PolylogN(MultiF s, MultiF x, int zI)  // z = Polylog(s, x)
		// Polylog(s, x) = Lerch(x, s, 1)
		{
			if (MultiFD.CompFF(s, MultiFD.fTwo) == 0)
			{
				DilogN(x, zI);			// Polylog(2, x) = Dilog(x);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			MultiF saveX = new MultiF(x.N);
			bo = saveX | x;
			LerchN(x, s, MultiFD.fOne, zI);
			XPCalLst.MulN(saveX, Common.nList[zI], zI);
			MultiFD.DoneF(saveX);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // DilogN

		public static void CosHN(MultiF x, int zI)  // z = CosH(x)
		// CosH(x) = (y + 1/y) / 2, where y = Exp(|x|)
		{
			int tI;
			Calc.SetToN(x, zI);
			if (Common.ZTestN(zI))
			{
				SetTo1N(1, zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "CosHN-tI", out tI);
			Common.nList[zI].S = false;
			ExpN(x, zI);
			if (!MultiID.muErr)
			{
				InvN(Common.nList[zI], tI);
				RAddN(Common.nList[tI], zI);
				RDiv1N(2, zI);
			}
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // CosHN

		public static void RCosHN(int zI) // z = CosH(z)
		{
			CosHN(Common.nList[zI], zI);
		}

		public static void TanHN(MultiF x, int zI)  // z = TanH(x)
		// TanH(x) = y / (y + 2), where y = Exp(2 * x) - 1
		// TanH(-x) = -TanH(x)
		{
			int tI;
			bool sign;
			bool saveRep;
			Calc.SetToN(x, zI);
			if (Common.ZTestN(zI) || (Common.nList[zI].C < (-Int32.MaxValue / 3 - Int16.MaxValue)) || MultiID.muErr) return;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "TanHN-tI", out tI);
			sign = Common.nList[zI].S;
			Common.nList[zI].S = false;
			saveRep = MultiID.muErrRep;
			MultiID.muErrRep = false;
			RMul1N(2, zI);
			RExpLN(zI);
			MultiID.muErrRep = saveRep;
			if (MultiID.muErr)		// TanH(+/- Large) = +/- 1
			{
				MultiID.muErr = false;
				Calc.ClearN(zI);
				Common.nList[zI].V[0] = 1;
			}
			else
			{
				Add1N(Common.nList[zI], 2, tI);
				RDivN(Common.nList[tI], zI);
			}
			Common.LDelete(tI);
			Common.nList[zI].S = sign;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // TanHN

		public static void RTanHN(int zI) // z = TanH(z)
		{
			TanHN(Common.nList[zI], zI);
		}

		public static void ASinHN(MultiF x, int zI)  // z = ASinH(x)
		// for (x >= 0.1 ASinH(x) = Ln(x + SqRt(1 + Sq(x)))
		// for (x <  0.1 ASinH(x) = LnL(x + Sq(x) / SqRt(1 + Sq(x)))
		// ASinH(-x) = -ASinH(x)
		{
			int tI;
			bool sign;
			bool saveRep;
			Calc.SetToN(x, zI);
			if (Common.ZTestN(zI) || (Common.nList[zI].C < (-Int32.MaxValue / 3 - Int16.MaxValue)) || MultiID.muErr) return;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "ASinHN-tI", out tI);
			sign = Common.nList[zI].S;
			Common.nList[zI].S = false;
			if ( (-Common.nList[zI].C > Common.nList[zI].N) ||
				((-Common.nList[zI].C == Common.nList[zI].N) &&
				(Common.nList[zI].V[Common.nList[zI].N-1] < MultiID.baseM / 10)) )
			{
				SqN(Common.nList[zI], tI);
				RAdd1N(1, tI);
				RSqRtN(tI);
				RAdd1N(1, tI);
				DivN(Common.nList[zI], Common.nList[tI], tI);
				RMulN(Common.nList[zI], tI);
				RAddN(Common.nList[tI], zI);
				Common.LDelete(tI);
				RLnLN(zI);
			}
			else
			{
				saveRep = MultiID.muErrRep;
				MultiID.muErrRep = false;
				SqN(Common.nList[zI], tI);
				MultiID.muErrRep = saveRep;
				if (MultiID.muErr)			// too big to square, so too big to add 1
				{
					MultiID.muErr = false;
					Calc.SetToII(zI, tI);
				}
				else
				{
					RAdd1N(1, tI);
					RSqRtN(tI);
				}
				RAddN(Common.nList[tI], zI);
				Common.LDelete(tI);
				RLnN(zI);
			}
			Common.nList[zI].S = sign;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ASinHN

		public static void RASinHN(int zI) // z = ASinH(z)
		{
			ASinHN(Common.nList[zI], zI);
		}

		public static void ACosHN(MultiF x, int zI)  // z = ACosH(x)
		// ACosH(x) = Ln(x + SqRt(Sq(x) - 1))
		// error if x < 1
		{
			int tI;
			if (MultiFD.ZTestF(x) || x.S || (-x.C >= x.N)) // if x < 1
			{
				MultiID.MuWriteErr("ACosH(x) error: x < 1.0");
				Calc.ClearN(zI);
				return;
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Calc.SetToN(x, zI);
			Common.InitNAdd(1, "ACosHN-tI", out tI);
			SqN(x, tI);
			if (!MultiID.muErr) // if no overflow
			{
				RAdd1N(-1, tI);
				RSqRtN(tI);
				RAddN(Common.nList[tI], zI);
				RLnN(zI);
			}
			else Calc.ClearN(zI);
			Common.LDelete(tI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ACosHN

		public static void RACosHN(int zI) // z = ACosH(z)
		{
			ACosHN(Common.nList[zI], zI);
		}

		public static void ATanHN(MultiF x, int zI)  // z = ATanH(x)
		// ATanH(x) = LnL(2 * x / (1 - x))
		// ATanH(-x) = -ATanH(x)
		// error if |x| >= 1
		{
			int tI;
			bool sign;
			Calc.SetToN(x, zI);
			if (Common.ZTestN(zI) || (Common.nList[zI].C < (-Int32.MaxValue / 3 - Int16.MaxValue))) return;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			Common.InitNAdd(1, "ATanHN-tI", out tI);
			sign = Common.nList[zI].S;
			Common.nList[zI].S = false;
			Add1N(Common.nList[zI], -1, tI);
			Common.ChSN(tI);
			if (Common.ZTestN(tI) || Common.nList[tI].S)
			{
				MultiID.MuWriteErr("ATanH(x) error: |x| >= 1.0");
				MultiFD.expanded = saveExp;
				Calc.fMC = saveFMC;
				MultiID.mMax = saveMMax;
				Common.LDelete(tI);
				Calc.ClearN(zI);
				return;
			}
			RDivN(Common.nList[tI], zI);
			Common.LDelete(tI);
			RMul1N(2, zI);
			RLnLN(zI);
			RDiv1N(2, zI);
			Common.nList[zI].S = sign;
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		} // ATanHN

		public static void RATanHN(int zI) // z = ATanH(z)
		{
			ATanHN(Common.nList[zI], zI);
		}

		public static void IntN(MultiF x, int zI)  // z = Integer part of x
		{
			Calc.SetToN(x, zI);
			MultiFD.RIntF(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void RIntN(int zI) // z = Integer part of z
		{
			IntN(Common.nList[zI], zI);
		}

		public static void FracN(MultiF x, int zI)  // z = Fractional part of x
		{
			Calc.SetToN(x, zI);
			MultiFD.RFracF(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void RFracN(int zI) // z = Fractional part of z
		{
			FracN(Common.nList[zI], zI);
		}

		public static void LopN(MultiF x, int zI) // z = x with one less super digit
		{
			MultiFD.LopF(x, Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void RLopN(int zI) // z = z with one less least significant "digit"
		{
			MultiFD.RLopF(Common.nList[zI]);
			Calc.NormN(zI);
		}

		public static void ToDegN(MultiF x, int zI) // z = x converted Rad to Degrees
		{
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			int pI = GetPiPtr();
			RMul1N(180, zI);
			RDivN(Common.nList[pI], zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RToDegN(int zI) // z = z converted Radians to Degrees
		{
			ToDegN(Common.nList[zI], zI);
		}

		public static void ToRadN(MultiF x, int zI)  // z = x converted Deg to Radians
		{
			int pI;
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			pI = GetPiPtr();
			RMulN(Common.nList[pI], zI);
			RDiv1N(180, zI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			Calc.NormN(zI);
		}

		public static void RToRadN(int zI) // z = z converted Deg to Radians
		{
			ToRadN(Common.nList[zI], zI);
		}

		public static void CalSFTn(int zI) // set number of digits in truncate in display
		{
			double db;
			MultiFD.GetDbInF(out db, 0, Int32.MaxValue, Common.nList[zI]);
			db = 1 + Math.Floor((db - 1) / MultiID.muDMax); // set for at least N decimal digits
			MultiFD.fTn = (int)db;
			Calc.CheckNMax();
		}

		public static void WriteFixedB(RichTextBox outB, StreamWriter outF, MultiF z, int tot, int width)
		{
			string ch;
			int d;
			int did;		// number of fractional digits output
			int dToD;		// digits to display
			int i;
			int k;
			int nOutF;		// number of significant digits in fraction output
			int nOutI;		// number of significant digits in integer output
			string st;
			string stD;		// formatted NOut
			MultiF t2 = new MultiF();
			k = 'A' - 10;
			nOutI = 0;
			nOutF = 0;
			did = 0;
			if (outB == null) MultiID.WriteTLn2(outB, outF, "", ref tot, width);
			MultiF t1 = new MultiF(z.N + 1);
			MultiFD.SetToF(z, t1); // t1 = z with one extra digit
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			t1.S = false;
			if (z.N + z.C > 0)
			{
				MultiFD.ReSizeF((int)(z.N + z.C), t2);
				MultiFD.IntF(t1, t2);
				MultiFD.RFracF(t1);
				if (MultiID.EQ1SI(1, t2))
				{
					MultiID.WriteT(outB, outF, "1", ref tot, width);
					nOutI = 1;
				}
				else
				{
					MultiID.WriteSI(outB, outF, t2, ref tot, width, false);
					nOutI = MultiID.nOutSI;
				}
				MultiFD.DoneF(t2);
			}
			else
			{
				MultiID.WriteT(outB, outF, "0", ref tot, width);
			}
			MultiID.WriteT(outB, outF, ".", ref tot, width);
			MultiFD.ReSizeF(1, t2);
			if (outB == null)
				dToD = 1 + (int)Math.Round((Calc.fMC * MultiID.muDMax) / Math.Log10(MultiID.baseO));
			else
				dToD = 1 + (int)Math.Round(((Calc.fMC - MultiFD.fTn) * MultiID.muDMax) / Math.Log10(MultiID.baseO));
			dToD = Math.Min(dToD, MultiID.maxDis);
			dToD = Math.Max(dToD, 2); // an extra digit is output, but not rounded
			do
			{
				MultiFD.RMul1F(MultiID.baseO, t1);
				MultiFD.IntF(t1, t2);
				MultiFD.RFracF(t1);
				MultiFD.Get1F(out d, t2);
				if (d < 10)
					ch = d.ToString();
				else
					ch = ((char)(d + k)).ToString();
				if (MultiID.muDpG > 0)
				{
					if ((did > 0) && ((did % MultiID.muDpG) == 0))
						MultiID.WriteT(outB, outF, ",", ref tot, width);
				}
				MultiID.WriteT(outB, outF, ch, ref tot, width);
				did++;
				if ((ch != "0") || ((nOutI + nOutF) > 0)) nOutF = nOutF + 1;
			}
			while (!MultiFD.ZTestF(t1) && !(nOutI + nOutF >= dToD));
			did = nOutI + nOutF;
			if (outF != null)
			{
				MultiID.Write(outB, outF, "\n");
				tot = 0;
				ch = "";
			}
			else
			{
				ch = " ";
			}
			stD = "";
			if (did > MultiID.muLenD)
				stD = nOutI.ToString() + "." + nOutF.ToString() + " Digits ";
			MultiID.WriteT(outB, outF, ch + "(" + stD + "Base " + MultiID.baseO + ")", ref tot, width);
			if (outF == null)
			{
				i = z.V[z.N-1].ToString().Length;
				st = (MultiID.baseM + z.V[0]).ToString();
				while (st[st.Length - 1] == '0') st = st.Remove(st.Length - 1, 1);
				i += st.Length - 1 - MultiID.muDMax;
				st = (i + (z.N - 1) * MultiID.muDMax).ToString();
				MultiID.WriteT(outB, outF, " [" + st + "]", ref tot, width);
			}
			MultiFD.DoneF(t2);
			MultiFD.DoneF(t1);
			return;
		} // WriteFixedB

		public static void WriteFB(RichTextBox outB, StreamWriter outF, MultiF z, ref int tot, int width)
		// output z as a line of text, tot = characters on line so far, z is not modified
		{
			int bI;				// natural log of baseO
			string ch;
			int d;
			decimal dE;			// exponent
			double db;
			int did;			// digits output
			int dToD;			// digits to display
			bool bFixed;
			int i;
			int k;
			int nI;
			string st;			// formatted number
			string stE;			// formatted exponent "E+xxx,xxx...x"
			MultiF t1 = new MultiF();	// temp
			MultiF t2 = new MultiF(1);
			int trunc;
			int min = 2;
			bFixed = (!MultiFD.scieN) && (z.N + z.C <= min) && (z.C >= -min);
			if (bFixed)
			{
				WriteFixedB(outB, outF, z, tot, 0); // format <= 99999999999999.0
				return;
			}
			if (outF != null) // always do full precision to file
				dToD = (int)Math.Round((Calc.fMC * MultiID.muDMax) / Math.Log10(MultiID.baseO)); // convert to baseO digits
			else
			{
				dToD = (int)Math.Round(((Calc.fMC - MultiFD.fTn) * MultiID.muDMax) / Math.Log10(MultiID.baseO));
				dToD = Math.Min(dToD, MultiID.maxDis);
				dToD = Math.Max(dToD, 2);	// needed or rounding factor will be wrong
			}
			bool saveExp = MultiFD.expanded;
			int saveFMC = Calc.fMC;
			int saveMMax = MultiID.mMax;
			Expand();
			db = MultiID.maxDis * Math.Log10(MultiID.baseO); // convert maxDis to Decimal digits
			trunc = (int)Math.Round(Math.Max(Calc.fMC - (db + 12) / 7, MultiFD.fTn));
			trunc = Math.Min(trunc, Calc.fMC - 2);
			if ((trunc > 0) && (outF == null))
				MultiFD.ReSizeF(Calc.fMC - trunc, t1);
			else
				MultiFD.ReSizeF(Calc.fMC, t1);
			MultiFD.SetToF(z, t1);	// rounds to desired digits
			t1.S = false;
			MultiF bF = new MultiF(Calc.fMC);	// baseO as a MultiF
			MultiFD.SetTo1F(MultiID.baseO, bF);
			Common.InitNAdd(6, "WriteFB-nI", out  nI);
			Common.InitNAdd(1, "WriteFB-bI", out bI);
			i = t1.V[t1.N-1].ToString().Length;
			dE = Decimal.Floor((( (decimal)t1.C + t1.N - 1) * MultiID.muDMax + i) / (decimal)Math.Log10(MultiID.baseO));
			MultiFD.SetToDeF(dE, Common.nList[nI]);
			MultiFD.SetTo1F(MultiID.baseO, Common.nList[bI]);
			if (dE < 0)
			{
				Common.nList[nI].S = false;
				RPowN(Common.nList[nI], bI);
				if (MultiID.muErr) goto WriteFBEnd;
				MultiFD.MulF(t1, Common.nList[bI], t1);
			}
			else if (dE > 0)
			{
				RPowN(Common.nList[nI], bI);
				if (MultiID.muErr) goto WriteFBEnd;
				MultiFD.DivF(t1, Common.nList[bI], t1);
			}
			while (MultiFD.CompFF(MultiFD.fOne, t1) < 0)	// while t1 < 1
			{
				MultiFD.RMul1F(MultiID.baseO, t1);
				dE--;
			}
			while (MultiFD.CompFF(bF, t1) >= 0)		// t1 >= baseO
			{
				MultiFD.RDiv1F(MultiID.baseO, t1);
				dE++;
			}
			MultiFD.ReSizeF(1, t2);
			if (MultiFD.fRound > 0)
			{
				SetTo1N(MultiID.baseO, bI);
				MultiFD.SetTo1F(-dToD + 1, t2);		// compute rounding factor (+1)
				RPowN(t2, bI);
				RDiv1N(2, bI);
				MultiFD.RAddF(Common.nList[bI], t1);
			}
			MultiFD.IntF(t1, t2);					// t2 = Integer part of t1
			MultiFD.Get1F(out d, t2);
			if (d == MultiID.baseO)
			{
				MultiFD.RDiv1F(MultiID.baseO, t1);
				dE++;
				MultiFD.IntF(t1, t2);				// t2 = Integer part of t1
				MultiFD.Get1F(out d, t2);
			}
			FormatStE(out stE, dE);
			if (outF != null)
				MultiID.WriteLn2(null, outF, " m.n " + stE + ", m.n ="); // line may > 80 ch
			MultiFD.RFracF(t1);						// t1 = Fractional part
			did = 0;
			k = 'A' - 10;
			st = "";
			Save1(ref st, ref did, d, k);			// first digit
			do
			{
				MultiFD.RMul1F(MultiID.baseO, t1);	// one digit at a time
				MultiFD.IntF(t1, t2);
				MultiFD.RFracF(t1);
				MultiFD.Get1F(out d, t2);
				Save1(ref st, ref did, d, k);
			}
			while ((did < dToD) && !MultiFD.ZTestF(t1));
			i = st.Length - 1;
			while (i > 1)				// built for speed, don't call WriteT too often
			{
				ch = st.Substring(i, 1);
				if (ch == "0")			// delete trailing zeros and don't count them
				{
					st = st.Substring(0, i);
					did--;
				}
				else if (ch == ",")		// delete trailing commas
					st = st.Substring(0, i);
				else
					break;				// exit while
				i--;
			}
			if (z.S) MultiID.WriteT(outB, outF, "-", ref tot, width);
			if (st[0] > '9') MultiID.WriteT(outB, outF, "0", ref tot, width);
			MultiID.WriteT(outB, outF, st.Substring(0, 1) + ".", ref tot, width);
			if (outF != null) MultiID.WriteTLn(null, outF, "", ref tot);
			MultiID.WriteT(outB, outF, st.Substring(1), ref tot, width);
			if (outF != null) MultiID.WriteTLn(null, outF, "", ref tot);
			st = "";
			if (did > MultiID.muLenD)
				st = did.ToString() + " Digits ";
			MultiID.WriteT(outB, outF, stE + " (" + st + "Base " + MultiID.baseO + ")", ref tot, width);
			if (outF == null)
			{
				i = z.V[z.N-1].ToString().Length;
				st = (MultiID.baseM + z.V[0]).ToString();
				while (st[st.Length - 1] == '0') st = st.Remove(st.Length - 1, 1);
				i += st.Length - 1 - MultiID.muDMax;
				st = (i + (z.N - 1) * MultiID.muDMax).ToString();
				MultiID.WriteT(outB, outF, " [" + st + "]", ref tot, width);
			}
WriteFBEnd:
			MultiFD.DoneF(t2);
			MultiFD.DoneF(bF);
			MultiFD.DoneF(t1);
			Common.LDelete(bI);
			Common.LDelete(nI);
			MultiFD.expanded = saveExp;
			Calc.fMC = saveFMC;
			MultiID.mMax = saveMMax;
			return;
		} // WriteFB

		private static void FormatStE(out string stE, decimal dE)
		// part of WriteFB
		{
			if (MultiID.baseO > 10) { stE = "`"; } else { stE = "E"; }
			if (dE >= 0) { stE = stE + "+"; } else { stE = stE + "-"; }
			string st = (Math.Abs(dE)).ToString();
			int k = st.Length;
			if ((k > MultiID.muDpG) && (MultiID.muDpG != 0))
			{
				int j = k + 1;
				int iL = (k - 1) / MultiID.muDpG;
				for (int i = 1; i <= iL; i++)
				{
					j -= MultiID.muDpG;
					st = st.Substring(0, j - 1) + "," + st.Substring(j - 1);
				}
			}
			stE += st;
			st = "";
			return;
		}

		private static void Save1(ref string st, ref int did, int d, int k)
		// part of WriteFB
		{
			string ch;
			if (MultiID.muDpG > 0)
			{
				if (((did - 1) % MultiID.muDpG) == 0)
				{
					if (did > 1) st = st + ",";
				}
			}
			if (d < 10)
				ch = d.ToString();
			else
				ch = ((char)(d + k)).ToString();
			st += ch;
			did++;
			return;
		}

		public static bool bo;		// dummy Boolean for bo = F | F assignment overload
		public static double[] c;	// 26 items at 1 through 26, constants for series for 1/gamma(x)
		public static int limFac;	// limit to use factorial function vice gamma function
	}
}

